from dataclasses import dataclass
from typing import TypedDict

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid import register_model
from lucid._tensor import Tensor

from lucid.models.utils import (
    ROIAlign,
    MultiScaleROIAlign,
    apply_deltas,
    bbox_to_delta,
    nms,
    iou,
    clip_boxes,
)

from lucid.models.vision.faster_rcnn import (
    _AnchorGenerator,
    _RegionProposalNetwork,
    _ResNetFPNBackbone,
)
from lucid.models.vision.resnet import resnet_50, resnet_101

__all__ = [
    "MaskRCNN",
    "MaskRCNNConfig",
    "mask_rcnn_resnet_50_fpn",
    "mask_rcnn_resnet_101_fpn",
]


@dataclass
class MaskRCNNConfig:
    backbone: nn.Module
    feat_channels: int
    num_classes: int
    use_fpn: bool = False
    anchor_sizes: tuple[int, ...] | list[int] = (128, 256, 512)
    aspect_ratios: tuple[float, ...] | list[float] = (0.5, 1.0, 2.0)
    anchor_stride: int = 16
    pool_size: tuple[int, int] | list[int] = (7, 7)
    hidden_dim: int = 4096
    dropout: float = 0.5
    mask_pool_size: tuple[int, int] | list[int] = (14, 14)
    mask_hidden_channels: int = 256
    mask_out_size: int = 28

    def __post_init__(self) -> None:
        if not isinstance(self.backbone, nn.Module):
            raise TypeError("backbone must be an nn.Module")
        if self.feat_channels <= 0:
            raise ValueError("feat_channels must be greater than 0")
        if self.num_classes <= 0:
            raise ValueError("num_classes must be greater than 0")
        if not isinstance(self.use_fpn, bool):
            raise TypeError("use_fpn must be a bool")

        self.anchor_sizes = tuple(self.anchor_sizes)
        self.aspect_ratios = tuple(self.aspect_ratios)
        self.pool_size = tuple(self.pool_size)
        self.mask_pool_size = tuple(self.mask_pool_size)

        if len(self.anchor_sizes) == 0 or any(size <= 0 for size in self.anchor_sizes):
            raise ValueError("anchor_sizes must contain at least one positive value")
        if len(self.aspect_ratios) == 0 or any(
            ratio <= 0 for ratio in self.aspect_ratios
        ):
            raise ValueError("aspect_ratios must contain at least one positive value")
        if self.anchor_stride <= 0:
            raise ValueError("anchor_stride must be greater than 0")
        if len(self.pool_size) != 2 or any(size <= 0 for size in self.pool_size):
            raise ValueError("pool_size must contain exactly two positive integers")
        if self.hidden_dim <= 0:
            raise ValueError("hidden_dim must be greater than 0")
        if self.dropout < 0 or self.dropout >= 1:
            raise ValueError("dropout must be in the range [0, 1)")
        if len(self.mask_pool_size) != 2 or any(
            size <= 0 for size in self.mask_pool_size
        ):
            raise ValueError(
                "mask_pool_size must contain exactly two positive integers"
            )
        if self.mask_pool_size[0] != self.mask_pool_size[1]:
            raise ValueError("mask_pool_size must be square")
        if self.mask_hidden_channels <= 0:
            raise ValueError("mask_hidden_channels must be greater than 0")
        if self.mask_out_size <= 0:
            raise ValueError("mask_out_size must be greater than 0")
        if self.mask_out_size != self.mask_pool_size[0] * 2:
            raise ValueError("mask_out_size must equal 2 * mask_pool_size[0]")


class _MaskHead(nn.Module):
    def __init__(self, in_channels: int, num_classes: int, hidden_channels: int = 256):
        super().__init__()
        self.conv1 = nn.Conv2d(in_channels, hidden_channels, kernel_size=3, padding=1)
        self.conv2 = nn.Conv2d(
            hidden_channels, hidden_channels, kernel_size=3, padding=1
        )
        self.conv3 = nn.Conv2d(
            hidden_channels, hidden_channels, kernel_size=3, padding=1
        )
        self.conv4 = nn.Conv2d(
            hidden_channels, hidden_channels, kernel_size=3, padding=1
        )
        self.deconv = nn.ConvTranspose2d(
            hidden_channels, hidden_channels, kernel_size=2, stride=2
        )
        self.mask_predictor = nn.Conv2d(hidden_channels, num_classes, kernel_size=1)

    def forward(self, x: Tensor) -> Tensor:
        x = F.relu(self.conv1(x))
        x = F.relu(self.conv2(x))
        x = F.relu(self.conv3(x))
        x = F.relu(self.conv4(x))
        x = F.relu(self.deconv(x))

        return self.mask_predictor(x)


class _MaskRCNNLoss(TypedDict):
    rpn_cls_loss: Tensor
    rpn_reg_loss: Tensor
    roi_cls_loss: Tensor
    roi_reg_loss: Tensor
    mask_loss: Tensor
    total_loss: Tensor


class MaskRCNN(nn.Module):
    def __init__(self, config: MaskRCNNConfig) -> None:
        super().__init__()
        self.config = config
        self.backbone = config.backbone
        self.anchor_generator = _AnchorGenerator(
            config.anchor_sizes, config.aspect_ratios, config.anchor_stride
        )
        self.rpn = _RegionProposalNetwork(config.feat_channels, self.anchor_generator)

        self.use_fpn = config.use_fpn
        self.roipool = (
            MultiScaleROIAlign(config.pool_size)
            if self.use_fpn
            else ROIAlign(config.pool_size)
        )
        self.mask_roipool = (
            MultiScaleROIAlign(config.mask_pool_size)
            if self.use_fpn
            else ROIAlign(config.mask_pool_size)
        )

        fc_in = config.feat_channels * config.pool_size[0] * config.pool_size[1]
        self.fc1 = nn.Linear(fc_in, config.hidden_dim)
        self.fc2 = nn.Linear(config.hidden_dim, config.hidden_dim)
        self.drop1 = nn.Dropout(config.dropout)
        self.drop2 = nn.Dropout(config.dropout)

        self.cls_score = nn.Linear(config.hidden_dim, config.num_classes)
        self.bbox_pred = nn.Linear(config.hidden_dim, config.num_classes * 4)
        self.mask_head = _MaskHead(
            config.feat_channels,
            config.num_classes,
            hidden_channels=config.mask_hidden_channels,
        )

        self.mask_out_size = config.mask_out_size
        self.mask_target_pool = ROIAlign((config.mask_out_size, config.mask_out_size))

        self.bbox_reg_means: nn.Buffer
        self.bbox_reg_stds: nn.Buffer
        self.register_buffer(
            "bbox_reg_means", Tensor([0.0, 0.0, 0.0, 0.0], dtype=lucid.Float32)
        )
        self.register_buffer(
            "bbox_reg_stds", Tensor([0.1, 0.1, 0.2, 0.2], dtype=lucid.Float32)
        )

    def _extract_features(self, images: Tensor) -> tuple[Tensor, Tensor | list[Tensor]]:
        feats = self.backbone(images)
        if self.use_fpn and isinstance(feats, (list, tuple)):
            return feats[0], feats
        return feats, feats

    def _flatten_proposals(
        self, proposals: list[Tensor], device: str
    ) -> tuple[Tensor, Tensor]:
        boxes_list: list[Tensor] = []
        idx_list: list[Tensor] = []
        for i, p in enumerate(proposals):
            if p.size == 0:
                continue

            boxes_list.append(p)
            idx_list.append(
                lucid.full((p.shape[0],), i, dtype=lucid.Int32, device=device)
            )

        if boxes_list:
            return lucid.concatenate(boxes_list, axis=0), lucid.concatenate(
                idx_list, axis=0
            )

        return (
            lucid.empty(0, 4, device=device),
            lucid.empty(0, dtype=lucid.Int32, device=device),
        )

    def roi_forward(
        self,
        feats: Tensor | list[Tensor],
        rois: Tensor,
        roi_idx: Tensor,
    ) -> tuple[Tensor, Tensor]:
        pooled = self.roipool(feats, rois, roi_idx)
        N = pooled.shape[0]
        if N == 0:
            return (
                lucid.empty(0, self.cls_score.weight.shape[0], device=rois.device),
                lucid.empty(0, self.bbox_pred.weight.shape[0], device=rois.device),
            )

        x = pooled.reshape(N, -1)
        x = F.relu(self.fc1(x))
        x = self.drop1(x)
        x = F.relu(self.fc2(x))
        x = self.drop2(x)

        cls_logits = self.cls_score(x)
        bbox_deltas = self.bbox_pred(x)

        return cls_logits, bbox_deltas

    def mask_forward(
        self, feats: Tensor | list[Tensor], rois: Tensor, roi_idx: Tensor
    ) -> Tensor:
        pooled = self.mask_roipool(feats, rois, roi_idx)
        if pooled.shape[0] == 0:
            return lucid.empty(
                0,
                self.mask_head.mask_predictor.weight.shape[0],
                self.mask_out_size,
                self.mask_out_size,
                device=rois.device,
            )
        return self.mask_head(pooled)

    def forward(
        self,
        images: Tensor,
        rois: Tensor | None = None,
        roi_idx: Tensor | None = None,
    ) -> tuple[Tensor, Tensor, Tensor]:
        _, H, W = images.shape[1:]
        feature_rpn, feature_roi = self._extract_features(images)

        if rois is None or roi_idx is None:
            proposals = self.rpn(feature_rpn, (H, W))
            proposals = [p.detach() for p in proposals]
            rois_px, roi_idx = self._flatten_proposals(proposals, images.device)

            rois = rois_px / lucid.Tensor([W, H, W, H], dtype=lucid.Float32)
            rois = rois.to(images.device)

        cls_logits, bbox_deltas = self.roi_forward(feature_roi, rois, roi_idx)
        mask_logits = self.mask_forward(feature_roi, rois, roi_idx)

        return cls_logits, bbox_deltas, mask_logits

    def _match_rois(
        self,
        proposals: list[Tensor],
        targets: list[dict[str, Tensor]],
    ) -> tuple[Tensor, Tensor, Tensor]:
        all_labels: list[Tensor] = []
        all_deltas: list[Tensor] = []
        all_gt_indices: list[Tensor] = []

        for i, props in enumerate(proposals):
            gt_boxes = targets[i]["boxes"].to(props.device)
            gt_labels = targets[i]["labels"].to(props.device)

            if props.size == 0:
                all_labels.append(
                    lucid.empty(0, dtype=lucid.Int32, device=props.device)
                )
                all_deltas.append(lucid.empty(0, 4, device=props.device))
                all_gt_indices.append(
                    lucid.empty(0, dtype=lucid.Int32, device=props.device)
                )
                continue

            if gt_boxes.size == 0:
                all_labels.append(
                    lucid.zeros(
                        (props.shape[0],), dtype=lucid.Int32, device=props.device
                    )
                )
                all_deltas.append(lucid.zeros((props.shape[0], 4), device=props.device))
                all_gt_indices.append(
                    lucid.zeros(
                        (props.shape[0],), dtype=lucid.Int32, device=props.device
                    )
                )
                continue

            ious = iou(props, gt_boxes)
            max_iou = lucid.max(ious, axis=1)
            argmax = lucid.argsort(ious, axis=1, descending=True)[:, 0]

            labels = gt_labels[argmax].astype(lucid.Int32)
            labels[max_iou < 0.5] = 0

            deltas = bbox_to_delta(props, gt_boxes[argmax])
            all_labels.append(labels)
            all_deltas.append(deltas)
            all_gt_indices.append(argmax.astype(lucid.Int32))

        return (
            lucid.concatenate(all_labels, axis=0),
            lucid.concatenate(all_deltas, axis=0),
            lucid.concatenate(all_gt_indices, axis=0),
        )

    @lucid.no_grad()
    def predict(
        self,
        images: Tensor,
        *,
        score_thresh: float = 0.05,
        nms_thresh: float = 0.5,
        top_k: int = 100,
    ) -> list[dict[str, Tensor]]:
        B, _, H, W = images.shape
        feature_rpn, feature_roi = self._extract_features(images)

        proposals = self.rpn(feature_rpn, (H, W))
        proposals = [p.detach() for p in proposals]

        rois_px, roi_idx = self._flatten_proposals(proposals, images.device)
        rois_norm = rois_px / lucid.Tensor([W, H, W, H], dtype=lucid.Float32)
        rois_norm = rois_norm.to(images.device)

        valid = (rois_norm[:, 2] > rois_norm[:, 0]) & (
            rois_norm[:, 3] > rois_norm[:, 1]
        )
        if lucid.sum(valid) == 0:
            empty_boxes = lucid.empty(0, 4, device=images.device)
            empty_scores = lucid.empty(0, device=images.device)
            empty_labels = lucid.empty(0, dtype=lucid.Int32, device=images.device)
            empty_masks = lucid.empty(
                0,
                self.mask_out_size,
                self.mask_out_size,
                device=images.device,
            )
            return [
                {
                    "boxes": empty_boxes,
                    "scores": empty_scores,
                    "labels": empty_labels,
                    "masks": empty_masks,
                }
                for _ in range(B)
            ]

        rois_px = rois_px[valid]
        rois_norm = rois_norm[valid]
        roi_idx = roi_idx[valid]

        cls_logits, bbox_deltas = self.roi_forward(feature_roi, rois_norm, roi_idx)
        scores = F.softmax(cls_logits, axis=1)

        num_classes = scores.shape[1]
        detections = [{"boxes": [], "scores": [], "labels": []} for _ in range(B)]

        for c in range(1, num_classes):
            cls_scores = scores[:, c]
            deltas_cls = bbox_deltas[:, c * 4 : (c + 1) * 4]
            deltas_cls = deltas_cls * self.bbox_reg_stds + self.bbox_reg_means

            boxes_all = apply_deltas(rois_px, deltas_cls)
            boxes_all = clip_boxes(boxes_all, (H, W))

            keep = cls_scores > score_thresh
            if lucid.sum(keep) == 0:
                continue

            cls_scores = cls_scores[keep]
            boxes = boxes_all[keep]
            img_ids = roi_idx[keep]

            for img_id in img_ids.unique():
                m = img_ids == img_id
                boxes_i = boxes[m]
                scores_i = cls_scores[m]

                keep_idx = nms(boxes_i, scores_i, nms_thresh)[:top_k]
                if keep_idx.size == 0:
                    continue

                det = detections[int(img_id.item())]
                det["boxes"].append(boxes_i[keep_idx])
                det["scores"].append(scores_i[keep_idx])
                det["labels"].append(
                    lucid.full(
                        (keep_idx.size,), c, dtype=lucid.Int32, device=images.device
                    )
                )

        for det in detections:
            if det["boxes"]:
                det["boxes"] = lucid.concatenate(det["boxes"], axis=0)
                det["scores"] = lucid.concatenate(det["scores"], axis=0)
                det["labels"] = lucid.concatenate(det["labels"], axis=0)
            else:
                det["boxes"] = lucid.empty(0, 4, device=images.device)
                det["scores"] = lucid.empty(0, device=images.device)
                det["labels"] = lucid.empty(0, dtype=lucid.Int32, device=images.device)

        all_boxes: list[Tensor] = []
        all_labels: list[Tensor] = []
        all_idx: list[Tensor] = []
        for i, det in enumerate(detections):
            n_det = det["boxes"].shape[0]
            if n_det == 0:
                continue
            all_boxes.append(det["boxes"])
            all_labels.append(det["labels"])
            all_idx.append(
                lucid.full((n_det,), i, dtype=lucid.Int32, device=images.device)
            )

        if all_boxes:
            det_rois_px = lucid.concatenate(all_boxes, axis=0)
            det_labels = lucid.concatenate(all_labels, axis=0)
            det_idx = lucid.concatenate(all_idx, axis=0)

            det_rois_norm = det_rois_px / lucid.Tensor(
                [W, H, W, H], dtype=lucid.Float32
            )
            det_rois_norm = det_rois_norm.to(images.device)

            mask_logits = self.mask_forward(feature_roi, det_rois_norm, det_idx)
            mask_probs: list[Tensor] = []
            for j, c in enumerate(det_labels):
                cls_idx = int(c.item())
                mask_probs.append(F.sigmoid(mask_logits[j, cls_idx]))
            mask_probs_all = lucid.stack(mask_probs, axis=0)

            offset = 0
            for det in detections:
                n_det = det["boxes"].shape[0]
                if n_det == 0:
                    det["masks"] = lucid.empty(
                        0,
                        self.mask_out_size,
                        self.mask_out_size,
                        device=images.device,
                    )
                else:
                    det["masks"] = mask_probs_all[offset : offset + n_det]
                offset += n_det

        else:
            for det in detections:
                det["masks"] = lucid.empty(
                    0,
                    self.mask_out_size,
                    self.mask_out_size,
                    device=images.device,
                )

        return detections

    def get_loss(
        self, images: Tensor, targets: list[dict[str, Tensor]]
    ) -> _MaskRCNNLoss:
        B, _, H, W = images.shape
        feature_rpn, feature_roi = self._extract_features(images)

        rpn_out = self.rpn.get_loss(feature_rpn, targets, (H, W))
        rpn_cls_loss = rpn_out["rpn_cls_loss"]
        rpn_reg_loss = rpn_out["rpn_reg_loss"]
        proposals = [p.detach() for p in rpn_out["proposals"]]

        rois_px, roi_idx = self._flatten_proposals(proposals, images.device)
        rois_norm = rois_px / lucid.Tensor([W, H, W, H], dtype=lucid.Float32)
        rois_norm = rois_norm.to(images.device)

        roi_labels, roi_reg_targets, roi_gt_indices = self._match_rois(
            proposals, targets
        )

        sampled_rois: list[Tensor] = []
        sampled_idx: list[Tensor] = []
        sampled_labels: list[Tensor] = []
        sampled_tgts: list[Tensor] = []
        sampled_gt_indices: list[Tensor] = []

        for i in range(B):
            mask_i = (roi_idx == i).nonzero().squeeze(axis=1)
            labels_i = roi_labels[mask_i]
            pos_i = (labels_i > 0).nonzero().squeeze(axis=1)
            neg_i = (labels_i == 0).nonzero().squeeze(axis=1)

            num_pos = min(pos_i.shape[0], 32)
            num_neg = min(neg_i.shape[0], 128 - num_pos)

            perm_pos = lucid.random.permutation(pos_i.shape[0], device=images.device)[
                :num_pos
            ]
            perm_neg = lucid.random.permutation(neg_i.shape[0], device=images.device)[
                :num_neg
            ]

            sel_idx = lucid.concatenate([pos_i[perm_pos], neg_i[perm_neg]], axis=0)
            sel = mask_i[sel_idx]

            sampled_rois.append(rois_norm[sel])
            sampled_idx.append(roi_idx[sel])
            sampled_labels.append(roi_labels[sel])
            sampled_tgts.append(roi_reg_targets[sel])
            sampled_gt_indices.append(roi_gt_indices[sel])

        sb_rois = lucid.concatenate(sampled_rois, axis=0)
        sb_idx = lucid.concatenate(sampled_idx, axis=0)
        sb_labels = lucid.concatenate(sampled_labels, axis=0)
        sb_tgts = lucid.concatenate(sampled_tgts, axis=0)
        sb_gt_idx = lucid.concatenate(sampled_gt_indices, axis=0)

        valid = (sb_rois[:, 2] > sb_rois[:, 0]) & (sb_rois[:, 3] > sb_rois[:, 1])
        if lucid.sum(valid) > 0:
            sb_rois = sb_rois[valid]
            sb_idx = sb_idx[valid]
            sb_labels = sb_labels[valid]
            sb_tgts = sb_tgts[valid]
            sb_gt_idx = sb_gt_idx[valid]
        else:
            sb_rois = lucid.empty(0, 4, device=images.device)

        if sb_rois.shape[0] == 0:
            zero = lucid.zeros((), dtype=lucid.Float32, device=images.device)
            total_loss = (rpn_cls_loss + rpn_reg_loss) / B
            return {
                "rpn_cls_loss": rpn_cls_loss,
                "rpn_reg_loss": rpn_reg_loss,
                "roi_cls_loss": zero,
                "roi_reg_loss": zero,
                "mask_loss": zero,
                "total_loss": total_loss,
            }

        cls_logits, bbox_deltas = self.roi_forward(feature_roi, sb_rois, sb_idx)
        roi_cls_loss = F.cross_entropy(cls_logits, sb_labels, reduction="mean")

        fg = (sb_labels > 0).nonzero().squeeze(axis=1)
        if fg.shape[0] > 0:
            preds: list[Tensor] = []
            for j, c in enumerate(sb_labels[fg]):
                start = int(c.item()) * 4
                preds.append(bbox_deltas[fg[j], start : start + 4])

            preds = lucid.stack(preds, axis=0)
            reg_tgt_fg = sb_tgts[fg]
            roi_reg_loss = F.huber_loss(preds, reg_tgt_fg, reduction="mean")

            fg_rois = sb_rois[fg]
            fg_idx = sb_idx[fg]
            fg_labels = sb_labels[fg]
            fg_gt_idx = sb_gt_idx[fg]

            mask_logits = self.mask_forward(feature_roi, fg_rois, fg_idx)
            pred_masks: list[Tensor] = []
            for j, c in enumerate(fg_labels):
                pred_masks.append(mask_logits[j, int(c.item())])
            pred_masks_t = lucid.stack(pred_masks, axis=0)

            target_masks: list[Tensor] = []
            for j in range(fg_rois.shape[0]):
                img_id = int(fg_idx[j].item())
                gt_id = int(fg_gt_idx[j].item())

                gt_masks_i = targets[img_id]["masks"].to(images.device)
                if gt_masks_i.ndim == 2:
                    gt_mask = gt_masks_i
                elif gt_masks_i.ndim == 3:
                    gt_mask = gt_masks_i[gt_id]
                elif gt_masks_i.ndim == 4:
                    gt_mask = gt_masks_i[gt_id, 0]
                else:
                    raise ValueError(
                        "targets[i]['masks'] must be (N,H,W), (N,1,H,W), or (H,W)"
                    )

                mask_img = (
                    gt_mask.astype(lucid.Float32).unsqueeze(axis=0).unsqueeze(axis=0)
                )
                roi_j = fg_rois[j].unsqueeze(axis=0)
                roi_j_idx = lucid.zeros((1,), dtype=lucid.Int32, device=images.device)

                pooled_tgt = self.mask_target_pool(mask_img, roi_j, roi_j_idx)
                if pooled_tgt.shape[0] == 0:
                    target_mask = lucid.zeros(
                        (self.mask_out_size, self.mask_out_size),
                        dtype=lucid.Float32,
                        device=images.device,
                    )
                else:
                    target_mask = pooled_tgt[0, 0]

                target_masks.append((target_mask >= 0.5).astype(lucid.Float32))

            target_masks_t = lucid.stack(target_masks, axis=0)
            mask_loss = F.binary_cross_entropy_with_logits(
                pred_masks_t, target_masks_t, reduction="mean"
            )

        else:
            roi_reg_loss = lucid.zeros((), dtype=lucid.Float32, device=images.device)
            mask_loss = lucid.zeros((), dtype=lucid.Float32, device=images.device)

        total_loss = (
            rpn_cls_loss + rpn_reg_loss + roi_cls_loss + roi_reg_loss + mask_loss
        ) / B
        return {
            "rpn_cls_loss": rpn_cls_loss,
            "rpn_reg_loss": rpn_reg_loss,
            "roi_cls_loss": roi_cls_loss,
            "roi_reg_loss": roi_reg_loss,
            "mask_loss": mask_loss,
            "total_loss": total_loss,
        }


@register_model
def mask_rcnn_resnet_50_fpn(
    num_classes: int = 21, backbone_num_classes: int = 1000, **kwargs
) -> MaskRCNN:
    config_kwargs = {
        "backbone": _ResNetFPNBackbone(resnet_50(backbone_num_classes)),
        "feat_channels": 256,
        "num_classes": num_classes,
        "use_fpn": True,
        "hidden_dim": 1024,
    }
    config_kwargs.update(kwargs)
    return MaskRCNN(MaskRCNNConfig(**config_kwargs))


@register_model
def mask_rcnn_resnet_101_fpn(
    num_classes: int = 21, backbone_num_classes: int = 1000, **kwargs
) -> MaskRCNN:
    config_kwargs = {
        "backbone": _ResNetFPNBackbone(resnet_101(backbone_num_classes)),
        "feat_channels": 256,
        "num_classes": num_classes,
        "use_fpn": True,
        "hidden_dim": 1024,
    }
    config_kwargs.update(kwargs)
    return MaskRCNN(MaskRCNNConfig(**config_kwargs))



import importlib
from collections.abc import Iterator
from typing import Any

import numpy as np
import torch

import lucid
import lucid.nn as nn
import lucid.optim as optim
import lucid.optim.lr_scheduler as lr_scheduler

from lucid.test.core import ModuleTorchBase


class _TopLevelBackbone(nn.Module):
    def __init__(self, num_classes: int = 10) -> None:
        super().__init__()

        self.identity = nn.Identity()
        self.stem = nn.Sequential(
            nn.Conv2d(
                in_channels=3,
                out_channels=8,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=True,
            ),
            nn.BatchNorm2d(
                num_features=8,
                eps=1e-5,
                momentum=0.1,
                affine=True,
                track_running_stats=False,
            ),
            nn.ReLU(),
        )
        self.blocks = nn.ModuleList(
            [
                nn.Sequential(
                    nn.Conv2d(
                        in_channels=8,
                        out_channels=8,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                        bias=True,
                    ),
                    nn.GroupNorm(num_groups=2, num_channels=8, eps=1e-5, affine=True),
                    nn.ReLU(),
                    nn.MaxPool2d(kernel_size=2, stride=2),
                ),
                nn.Sequential(
                    nn.Conv2d(
                        in_channels=8,
                        out_channels=8,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                        bias=True,
                    ),
                    nn.GELU(),
                ),
            ]
        )
        self.layer_norm = nn.LayerNorm((8, 16, 16), eps=1e-5)
        self.dropout = nn.Dropout2d(p=0.0)
        self.pool = nn.AdaptiveAvgPool2d(output_size=(4, 4))
        self.flatten = nn.Flatten(start_axis=1, end_axis=-1)

        self.classifier = nn.Sequential(
            nn.Linear(in_features=128, out_features=64, bias=True),
            nn.LayerNorm(64, eps=1e-5),
            nn.Dropout(p=0.0),
            nn.Linear(in_features=64, out_features=num_classes, bias=True),
        )

    def forward(self, input_: Any) -> Any:
        x = self.identity(input_)
        x = self.stem(x)

        for block in self.blocks:
            x = block(x)

        x = self.layer_norm(x)
        x = self.dropout(x)
        x = self.pool(x)

        x = self.flatten(x)
        x = self.classifier(x)

        return x


class _TorchTopLevelBackbone(torch.nn.Module):
    def __init__(self, num_classes: int = 10) -> None:
        super().__init__()

        self.identity = torch.nn.Identity()
        self.stem = torch.nn.Sequential(
            torch.nn.Conv2d(
                in_channels=3,
                out_channels=8,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=True,
            ),
            torch.nn.BatchNorm2d(
                num_features=8,
                eps=1e-5,
                momentum=0.1,
                affine=True,
                track_running_stats=False,
            ),
            torch.nn.ReLU(),
        )
        self.blocks = torch.nn.ModuleList(
            [
                torch.nn.Sequential(
                    torch.nn.Conv2d(
                        in_channels=8,
                        out_channels=8,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                        bias=True,
                    ),
                    torch.nn.GroupNorm(num_groups=2, num_channels=8, eps=1e-5),
                    torch.nn.ReLU(),
                    torch.nn.MaxPool2d(kernel_size=2, stride=2),
                ),
                torch.nn.Sequential(
                    torch.nn.Conv2d(
                        in_channels=8,
                        out_channels=8,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                        bias=True,
                    ),
                    torch.nn.GELU(approximate="tanh"),
                ),
            ]
        )
        self.layer_norm = torch.nn.LayerNorm((8, 16, 16), eps=1e-5)
        self.dropout = torch.nn.Dropout2d(p=0.0)
        self.pool = torch.nn.AdaptiveAvgPool2d((4, 4))
        self.flatten = torch.nn.Flatten(start_dim=1, end_dim=-1)

        self.classifier = torch.nn.Sequential(
            torch.nn.Linear(in_features=128, out_features=64, bias=True),
            torch.nn.LayerNorm(64, eps=1e-5),
            torch.nn.Dropout(p=0.0),
            torch.nn.Linear(in_features=64, out_features=num_classes, bias=True),
        )

    def forward(self, input_: torch.Tensor) -> torch.Tensor:
        x = self.identity(input_)
        x = self.stem(x)

        for block in self.blocks:
            x = block(x)

        x = self.layer_norm(x)
        x = self.dropout(x)
        x = self.pool(x)

        x = self.flatten(x)
        x = self.classifier(x)

        return x


class TestTopLevelModuleTraining(ModuleTorchBase):
    random_seed: int | None = 20260305
    batch_size: int = 8
    num_classes: int = 10
    steps: int = 64
    lr: float = 5e-3

    torch_seed_offset: int = 101
    scheduler_step_size: int = 2
    scheduler_gamma: float = 0.5
    compare_scheduler: bool = True
    scheduler_last_epoch: int = -1

    np_dtype: Any = np.float32
    rtol: float = 5e-5
    atol: float = 1e-6
    param_rtol: float = 5e-4
    param_atol: float = 1e-5
    param_grad_skip_threshold: float = 1e-7

    def tensor_op_devices(self) -> tuple[str, ...]:
        return ("cpu", "gpu")

    @staticmethod
    def _torch_device(device: str) -> str:
        return "mps" if device == "gpu" else device

    def _set_seeds(self, seed: int, device: str) -> None:
        random = importlib.import_module("lucid.random")
        random.seed(seed)

        torch_seed = seed + self.torch_seed_offset
        torch.manual_seed(torch_seed)
        torch_device = self._torch_device(device)

        if torch_device == "mps" and hasattr(torch, "mps"):
            torch.mps.manual_seed(torch_seed)

        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(torch_seed)

    @staticmethod
    def _sync_model_state(
        lucid_model: nn.Module,
        torch_model: torch.nn.Module,
    ) -> None:
        lucid_params = tuple(lucid_model.parameters())
        torch_params = tuple(torch_model.parameters())

        if len(lucid_params) != len(torch_params):
            raise AssertionError(
                "Parameter count mismatch between lucid and torch models: "
                f"lucid={len(lucid_params)} torch={len(torch_params)}"
            )
        for lucid_param, torch_param in zip(lucid_params, torch_params):
            torch_param.data = torch.as_tensor(
                np.array(lucid_param.data),
                device=torch_param.device,
                dtype=torch_param.dtype,
            )

        lucid_buffers = tuple(lucid_model.buffers())
        torch_buffers = tuple(torch_model.buffers())

        if len(lucid_buffers) != len(torch_buffers):
            raise AssertionError(
                "Buffer count mismatch between lucid and torch models: "
                f"lucid={len(lucid_buffers)} torch={len(torch_buffers)}"
            )

        for lucid_buffer, torch_buffer in zip(lucid_buffers, torch_buffers):
            if lucid_buffer is None and torch_buffer is None:
                continue

            if lucid_buffer is None or torch_buffer is None:
                raise AssertionError("Buffer presence mismatch between models.")

            torch_buffer.data = torch.as_tensor(
                np.array(lucid_buffer.data),
                device=torch_buffer.device,
                dtype=torch_buffer.dtype,
            )

    @staticmethod
    def _iter_param_pairs(
        lucid_model: nn.Module,
        torch_model: torch.nn.Module,
    ) -> Iterator[tuple[Any, Any]]:
        lucid_params = tuple(lucid_model.parameters())
        torch_params = tuple(torch_model.parameters())

        if len(lucid_params) != len(torch_params):
            raise AssertionError(
                "Parameter count mismatch between lucid and torch models: "
                f"lucid={len(lucid_params)} torch={len(torch_params)}"
            )

        for lucid_param, torch_param in zip(lucid_params, torch_params):
            yield lucid_param, torch_param

    def _build_data(self, device: str) -> tuple[Any, Any, torch.Tensor, torch.Tensor]:
        rng = np.random.RandomState(self.random_seed + self.torch_seed_offset)
        x_np = rng.standard_normal((self.batch_size, 3, 32, 32)).astype(self.np_dtype)
        y_np = rng.standard_normal((self.batch_size, self.num_classes)).astype(
            self.np_dtype
        )

        torch_device = self._torch_device(device)
        x_torch = torch.tensor(
            x_np,
            dtype=torch.float32,
            device=torch_device,
            requires_grad=False,
        )
        y_torch = torch.tensor(
            y_np,
            dtype=torch.float32,
            device=torch_device,
            requires_grad=False,
        )

        x_lucid = self.tensor(
            x_np, requires_grad=False, dtype=lucid.Float32, device=device
        )
        y_lucid = self.tensor(
            y_np, requires_grad=False, dtype=lucid.Float32, device=device
        )

        return x_lucid, y_lucid, x_torch, y_torch

    @staticmethod
    def _to_numpy(value: Any) -> Any:
        if hasattr(value, "detach"):
            return value.detach().cpu().numpy()

        if hasattr(value, "data"):
            return np.array(value.data)

        return np.array(value)

    def test_module_op_forward_backward(self) -> None:
        return

    def test_modules_training_stepwise_match_torch(self) -> None:
        for device in self.tensor_op_devices():
            if not self._is_device_supported(device):
                continue

            self._set_seeds(seed=int(self.random_seed), device=device)
            x_lucid, y_lucid, x_torch, y_torch = self._build_data(device=device)

            lucid_model = _TopLevelBackbone(num_classes=self.num_classes)
            torch_model = _TorchTopLevelBackbone(num_classes=self.num_classes).to(
                self._torch_device(device)
            )
            lucid_model.to(device)

            self._sync_model_state(lucid_model, torch_model)
            lucid_model.train(True)
            torch_model.train(True)

            lucid_loss_fn = nn.MSELoss(reduction="mean")
            torch_loss_fn = torch.nn.MSELoss(reduction="mean")

            lucid_optimizer = optim.Adam(
                lucid_model.parameters(),
                lr=self.lr,
            )
            torch_optimizer = torch.optim.Adam(
                torch_model.parameters(),
                lr=self.lr,
            )
            lucid_scheduler = lr_scheduler.StepLR(
                optimizer=lucid_optimizer,
                step_size=self.scheduler_step_size,
                gamma=self.scheduler_gamma,
                last_epoch=self.scheduler_last_epoch,
            )

            for group in torch_optimizer.param_groups:
                group.setdefault("initial_lr", group["lr"])

            torch_scheduler = torch.optim.lr_scheduler.StepLR(
                torch_optimizer,
                step_size=self.scheduler_step_size,
                gamma=self.scheduler_gamma,
                last_epoch=self.scheduler_last_epoch,
            )

            for step in range(self.steps):
                lucid_optimizer.zero_grad()
                torch_optimizer.zero_grad()

                lucid_out = lucid_model(x_lucid)
                torch_out = torch_model(x_torch)

                assert_has = f"[step={step}, device={device}]"
                self.assert_tensor_allclose(
                    lucid_out,
                    torch_out,
                    rtol=self.rtol,
                    atol=self.atol,
                )

                lucid_loss = lucid_loss_fn(lucid_out, y_lucid)
                torch_loss = torch_loss_fn(torch_out, y_torch)

                self.assert_tensor_allclose(
                    lucid_loss,
                    torch_loss,
                    rtol=self.rtol,
                    atol=self.atol,
                )

                lucid_loss.backward()
                torch_loss.backward()

                lucid_param_grads: list[Any] = []
                torch_param_grads: list[Any] = []
                for lucid_param, torch_param in self._iter_param_pairs(
                    lucid_model, torch_model
                ):
                    if lucid_param.grad is None:
                        assert torch_param.grad is None, (
                            f"grad mismatch on {assert_has}: "
                            "torch has grad while lucid does not."
                        )
                        continue

                    assert (
                        torch_param.grad is not None
                    ), f"missing torch grad on {assert_has}"

                    self.assert_tensor_allclose(
                        lucid_param.grad,
                        torch_param.grad.detach().cpu().numpy(),
                        rtol=self.param_rtol,
                        atol=self.param_atol,
                    )
                    lucid_param_grads.append(lucid_param.grad)
                    torch_param_grads.append(torch_param.grad)

                assert lucid_param_grads, f"no params found for comparison {assert_has}"
                assert torch_param_grads, f"no params found for comparison {assert_has}"

                for g in lucid_param_grads:
                    assert np.isfinite(np.array(g)).all()
                for g in torch_param_grads:
                    assert np.isfinite(self._to_numpy(g)).all()

                lucid_optimizer.step()
                torch_optimizer.step()
                lucid_scheduler.step()
                torch_scheduler.step()

                if self.compare_scheduler:
                    self.assert_tensor_allclose(
                        [pg["lr"] for pg in lucid_optimizer.param_groups],
                        [pg["lr"] for pg in torch_optimizer.param_groups],
                        rtol=self.rtol,
                        atol=self.atol,
                    )

                for lucid_param, torch_param in self._iter_param_pairs(
                    lucid_model, torch_model
                ):
                    torch_grad = torch_param.grad
                    if torch_grad is not None:
                        torch_grad_arr = self._to_numpy(torch_grad)
                        if (
                            np.max(np.abs(torch_grad_arr))
                            < self.param_grad_skip_threshold
                        ):
                            continue

                    self.assert_tensor_allclose(
                        lucid_param,
                        torch_param.data,
                        rtol=self.param_rtol,
                        atol=self.param_atol,
                    )



import lucid
import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor

from lucid.models.text.bert import _wrappers as bert

from ._model import RoFormer, RoFormerConfig

__all__ = [
    "RoFormerForMaskedLM",
    "RoFormerForSequenceClassification",
    "RoFormerForTokenClassification",
    "RoFormerForMultipleChoice",
    "RoFormerForQuestionAnswering",
]


@register_model
class RoFormerForMaskedLM(bert.BERTForMaskedLM):
    def __init__(self, config: RoFormerConfig) -> None:
        super().__init__(config)
        self.bert = RoFormer(config)
        self.cls.apply(self.bert._init_weights)
        self.bert.set_output_embeddings(self.cls.predictions.decoder)


@register_model
class RoFormerForSequenceClassification(bert.BERTForSequenceClassification):
    def __init__(self, config: RoFormerConfig, num_labels: int = 2) -> None:
        super().__init__(config, num_labels=num_labels)
        self.bert = RoFormer(config)
        self.classifier.apply(self.bert._init_weights)


@register_model
class RoFormerForTokenClassification(bert.BERTForTokenClassification):
    def __init__(self, config: RoFormerConfig, num_labels: int = 2) -> None:
        super().__init__(config, num_labels=num_labels)
        self.bert = RoFormer(config)
        self.classifier.apply(self.bert._init_weights)


@register_model
class RoFormerForMultipleChoice(bert._BERTTaskWrapperMixin, nn.Module):
    def __init__(self, config: RoFormerConfig) -> None:
        super().__init__()
        self.config = config
        self.bert = RoFormer(config)

        classifier_dropout = config.classifier_dropout
        if classifier_dropout is None:
            classifier_dropout = config.hidden_dropout_prob

        self.dropout = nn.Dropout(classifier_dropout)
        self.classifier = nn.Linear(config.hidden_size, 1)
        self.classifier.apply(self.bert._init_weights)

    def forward(
        self,
        input_ids: lucid.LongTensor | None = None,
        attention_mask: Tensor | None = None,
        token_type_ids: lucid.LongTensor | None = None,
        position_ids: lucid.LongTensor | None = None,
        inputs_embeds: lucid.FloatTensor | None = None,
    ) -> Tensor:
        if input_ids is not None:
            batch_size, num_choices = input_ids.shape[:2]
            flat_input_ids = input_ids.reshape(-1, input_ids.shape[-1])
        elif inputs_embeds is not None:
            batch_size, num_choices = inputs_embeds.shape[:2]
            flat_input_ids = None
        else:
            raise ValueError("Either input_ids or inputs_embeds must be provided.")

        flat_attention_mask = (
            attention_mask.reshape(-1, attention_mask.shape[-1])
            if attention_mask is not None
            else None
        )
        flat_token_type_ids = (
            token_type_ids.reshape(-1, token_type_ids.shape[-1])
            if token_type_ids is not None
            else None
        )
        flat_position_ids = (
            position_ids.reshape(-1, position_ids.shape[-1])
            if position_ids is not None
            else None
        )
        flat_inputs_embeds = (
            inputs_embeds.reshape(-1, *inputs_embeds.shape[-2:])
            if inputs_embeds is not None
            else None
        )

        _, pooled_output = self.bert(
            **self._encoder_kwargs(
                input_ids=flat_input_ids,
                attention_mask=flat_attention_mask,
                token_type_ids=flat_token_type_ids,
                position_ids=flat_position_ids,
                inputs_embeds=flat_inputs_embeds,
            )
        )
        if pooled_output is None:
            raise RuntimeError("RoFormerForMultipleChoice requires pooled output.")

        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)
        return logits.reshape(batch_size, num_choices)

    def _forward_logits(
        self,
        input_ids: lucid.LongTensor | None = None,
        attention_mask: Tensor | None = None,
        token_type_ids: lucid.LongTensor | None = None,
        position_ids: lucid.LongTensor | None = None,
        inputs_embeds: lucid.FloatTensor | None = None,
    ) -> Tensor:
        return self.forward(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            inputs_embeds=inputs_embeds,
        )

    def get_loss(
        self,
        labels: Tensor,
        input_ids: lucid.LongTensor | None = None,
        attention_mask: Tensor | None = None,
        token_type_ids: lucid.LongTensor | None = None,
        position_ids: lucid.LongTensor | None = None,
        inputs_embeds: lucid.FloatTensor | None = None,
        *,
        reduction: str | None = "mean",
    ) -> Tensor:
        logits = self._forward_logits(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            inputs_embeds=inputs_embeds,
        )
        return bert._classification_ce_loss(logits, labels, reduction=reduction)

    def predict_labels(
        self,
        input_ids: lucid.LongTensor | None = None,
        attention_mask: Tensor | None = None,
        token_type_ids: lucid.LongTensor | None = None,
        position_ids: lucid.LongTensor | None = None,
        inputs_embeds: lucid.FloatTensor | None = None,
    ) -> Tensor:
        logits = self._forward_logits(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            inputs_embeds=inputs_embeds,
        )
        return bert._predict_labels_from_logits(logits)

    def predict_proba(
        self,
        input_ids: lucid.LongTensor | None = None,
        attention_mask: Tensor | None = None,
        token_type_ids: lucid.LongTensor | None = None,
        position_ids: lucid.LongTensor | None = None,
        inputs_embeds: lucid.FloatTensor | None = None,
    ) -> Tensor:
        logits = self._forward_logits(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            inputs_embeds=inputs_embeds,
        )
        return bert._predict_proba_from_logits(logits)

    def get_accuracy(
        self,
        labels: Tensor,
        input_ids: lucid.LongTensor | None = None,
        attention_mask: Tensor | None = None,
        token_type_ids: lucid.LongTensor | None = None,
        position_ids: lucid.LongTensor | None = None,
        inputs_embeds: lucid.FloatTensor | None = None,
    ) -> Tensor:
        preds = self.predict_labels(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            inputs_embeds=inputs_embeds,
        )
        return bert._classification_accuracy(preds, labels)


@register_model
class RoFormerForQuestionAnswering(bert.BERTForQuestionAnswering):
    def __init__(self, config: RoFormerConfig) -> None:
        super().__init__(config)
        self.bert = RoFormer(config)
        self.qa_outputs = nn.Linear(config.hidden_size, 2)
        self.qa_outputs.apply(self.bert._init_weights)



from typing import Any
from abc import ABC, abstractmethod

from lucid.optim import Optimizer


class LRScheduler(ABC):
    def __init__(
        self, optimizer: Optimizer, last_epoch: int = -1, verbose: bool = False
    ) -> None:
        if not hasattr(optimizer, "param_groups"):
            raise TypeError(f"{type(optimizer).__name__} is not a valid optimizer.")

        self.optimizer = optimizer
        self.last_epoch = last_epoch
        self.verbose = verbose
        self.base_lrs: list[float] = [float(g["lr"]) for g in optimizer.param_groups]

        self._step_count = 0
        self._last_lr: list[float] = [float(g["lr"]) for g in optimizer.param_groups]

    @abstractmethod
    def get_lr(self) -> list[float]:
        raise NotImplementedError

    def step(self, epoch: int | None = None) -> None:
        if epoch is None:
            self.last_epoch += 1
        else:
            self.last_epoch = int(epoch)
        self._step_count += 1

        new_lrs = self.get_lr()
        if len(new_lrs) != len(self.optimizer.param_groups):
            raise ValueError(
                f"get_lr returned {len(new_lrs)} values, "
                f"but optimizer has {len(self.optimizer.param_groups)} param groups."
            )

        for group, lr in zip(self.optimizer.param_groups, new_lrs):
            group["lr"] = float(lr)

        self._last_lr = [float(g["lr"]) for g in self.optimizer.param_groups]

        if self.verbose:
            print(
                f"Epoch {self.last_epoch}: setting learning rates to {self._last_lr}."
            )

    def state_dict(self) -> dict[str, Any]:
        return {
            "last_epoch": int(self.last_epoch),
            "base_lrs": [float(x) for x in self.base_lrs],
            "_step_count": int(self._step_count),
            "_last_lr": [float(x) for x in self._last_lr],
            "_group_count": len(self.optimizer.param_groups),
        }

    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        required = {"last_epoch", "base_lrs", "_step_count", "_last_lr"}
        missing = required - set(state_dict)
        if missing:
            raise KeyError(f"Missing keys in scheduler state_dict: {missing}")

        saved_group_count = int(
            state_dict.get("_group_count", len(state_dict["_last_lr"]))
        )
        current_group_count = len(self.optimizer.param_groups)
        if saved_group_count != current_group_count:
            raise ValueError(
                "Cannot load scheduler state: param group count mismatch "
                f"(saved={saved_group_count}, current={current_group_count})."
            )

        self.last_epoch = int(state_dict["last_epoch"])
        self.base_lrs = [float(x) for x in state_dict["base_lrs"]]
        self._step_count = int(state_dict["_step_count"])
        self._last_lr = [float(x) for x in state_dict["_last_lr"]]

        for group, lr in zip(self.optimizer.param_groups, self._last_lr):
            group["lr"] = float(lr)

    @property
    def last_lr(self) -> list[float]:
        return list(self._last_lr)



from __future__ import annotations
import typing

__all__: list[str] = list()

class _C_FuncOpSpec:
    device: str
    has_gradient: bool
    def __init__(self) -> None: ...
    @property
    def n_in(self) -> int | None: ...
    @n_in.setter
    def n_in(self, arg0: typing.SupportsInt | None) -> None: ...
    @property
    def n_ret(self) -> int | None: ...
    @n_ret.setter
    def n_ret(self, arg0: typing.SupportsInt | None) -> None: ...

def _C_func_op(
    op_self: typing.Any,
    forward_func: typing.Any,
    args: tuple,
    kwargs: dict,
    spec: _C_FuncOpSpec,
) -> typing.Any: ...
def _C_func_op_raw(
    op_self: typing.Any,
    forward_func: typing.Any,
    args: tuple,
    kwargs: dict,
    n_in: typing.SupportsInt | None = None,
    n_ret: typing.SupportsInt | None = None,
    has_gradient: bool = True,
    device: str = "cpu",
) -> typing.Any: ...



import functools
from types import ModuleType
from typing import Sequence

import numpy as np

from lucid._backend.core import Operation, func_op, _FuncOpReturnType, _GradType
from lucid._backend.metal import mx

from lucid._tensor import Tensor
from lucid.types import _DeviceType


def _norm_axes(ndim: int, normalized_shape: Sequence[int]) -> tuple[int, ...]:
    return tuple(range(ndim - len(normalized_shape), ndim))


def _broadcast_shape(ndim: int, normalized_shape: Sequence[int]) -> tuple[int, ...]:
    return (1,) * (ndim - len(normalized_shape)) + tuple(normalized_shape)


def _clone_array(arr):
    if isinstance(arr, np.ndarray):
        return arr.copy()
    return mx.array(arr)


class layer_norm_kernel(Operation):
    def __init__(
        self,
        normalized_shape: Sequence[int],
        eps: float = 1e-5,
        has_weight: bool = True,
        has_bias: bool = True,
    ) -> None:
        super().__init__()
        self.normalized_shape = tuple(int(v) for v in normalized_shape)
        self.eps = float(eps)
        self.has_weight = bool(has_weight)
        self.has_bias = bool(has_bias)

        self._xhat = None
        self._rstd = None
        self._norm_axes = None
        self._n = None

    def clear(self) -> None:
        super().clear()
        self._xhat = None
        self._rstd = None
        self._norm_axes = None
        self._n = None

    @func_op(n_in=3, n_ret=1, device="cpu")
    def cpu(self, a: Tensor, w: Tensor, b: Tensor) -> _FuncOpReturnType:
        return self._forward(a, w, b, lib_=np, device="cpu")

    @func_op(n_in=3, n_ret=1, device="gpu")
    def gpu(self, a: Tensor, w: Tensor, b: Tensor) -> _FuncOpReturnType:
        return self._forward(a, w, b, lib_=mx, device="gpu")

    def _forward(
        self,
        a: Tensor,
        w: Tensor,
        b: Tensor,
        lib_: ModuleType,
        device: _DeviceType,
    ) -> _FuncOpReturnType:
        norm_axes = _norm_axes(a.ndim, self.normalized_shape)
        n = int(np.prod(self.normalized_shape))
        mean = lib_.mean(a.data, axis=norm_axes, keepdims=True)
        var = lib_.var(a.data, axis=norm_axes, keepdims=True)
        rstd = 1.0 / lib_.sqrt(var + self.eps)
        xhat = (a.data - mean) * rstd

        out = xhat
        if self.has_weight:
            out = out * w.data.reshape(_broadcast_shape(a.ndim, self.normalized_shape))
        if self.has_bias:
            out = out + b.data.reshape(_broadcast_shape(a.ndim, self.normalized_shape))

        self._xhat = xhat
        self._rstd = rstd
        self._norm_axes = norm_axes
        self._n = n

        self.result = Tensor(out, device=device)
        return self.result, functools.partial(self.__grad__, a=a, w=w, lib_=lib_)

    def __grad__(self, a: Tensor, w: Tensor, lib_: ModuleType) -> _GradType:
        if self.result is None or self.result.grad is None:
            raise RuntimeError("layer_norm backward called before forward.")

        if self._xhat is None or self._rstd is None or self._norm_axes is None:
            raise RuntimeError("layer_norm cached data missing.")

        dy = self.result.grad
        xhat = self._xhat
        rstd = self._rstd
        norm_axes = self._norm_axes
        n = self._n if self._n is not None else int(np.prod(self.normalized_shape))

        if self.has_weight:
            w_broadcast = w.data.reshape(
                _broadcast_shape(a.ndim, self.normalized_shape)
            )
            dyw = dy * w_broadcast
        else:
            dyw = dy

        sum1 = lib_.sum(dyw, axis=norm_axes, keepdims=True)
        sum2 = lib_.sum(dyw * xhat, axis=norm_axes, keepdims=True)

        dx = (1.0 / n) * rstd * (n * dyw - sum1 - xhat * sum2)

        reduce_axes = tuple(range(0, a.ndim - len(self.normalized_shape)))
        if reduce_axes:
            dweight = lib_.sum(dy * xhat, axis=reduce_axes)
            dbias = lib_.sum(dy, axis=reduce_axes)
        else:
            dweight = dy * xhat
            dbias = dy

        return dx, dweight, dbias


class batch_norm_kernel(Operation):
    def __init__(
        self,
        eps: float = 1e-5,
        momentum: float = 0.1,
        training: bool = True,
        has_running: bool = True,
        has_weight: bool = True,
        has_bias: bool = True,
    ) -> None:
        super().__init__()
        self.eps = float(eps)
        self.momentum = float(momentum)
        self.training = bool(training)
        self.has_running = bool(has_running)
        self.has_weight = bool(has_weight)
        self.has_bias = bool(has_bias)

        self._xhat = None
        self._rstd = None
        self._axes = None
        self._m = None
        self._use_batch_stats = None

    def clear(self) -> None:
        super().clear()
        self._xhat = None
        self._rstd = None
        self._axes = None
        self._m = None
        self._use_batch_stats = None

    @func_op(n_in=5, n_ret=1, device="cpu")
    def cpu(
        self, a: Tensor, running_mean: Tensor, running_var: Tensor, w: Tensor, b: Tensor
    ) -> _FuncOpReturnType:
        return self._forward(a, running_mean, running_var, w, b, lib_=np, device="cpu")

    @func_op(n_in=5, n_ret=1, device="gpu")
    def gpu(
        self, a: Tensor, running_mean: Tensor, running_var: Tensor, w: Tensor, b: Tensor
    ) -> _FuncOpReturnType:
        return self._forward(a, running_mean, running_var, w, b, lib_=mx, device="gpu")

    def _forward(
        self,
        a: Tensor,
        running_mean: Tensor,
        running_var: Tensor,
        w: Tensor,
        b: Tensor,
        lib_: ModuleType,
        device: _DeviceType,
    ) -> _FuncOpReturnType:
        axes = (0,) + tuple(range(2, a.ndim))
        m = int(np.prod([a.shape[i] for i in axes]))
        use_batch_stats = self.training or not self.has_running

        if use_batch_stats:
            mean = lib_.mean(a.data, axis=axes, keepdims=True)
            var = lib_.var(a.data, axis=axes, keepdims=True)

            if self.training and self.has_running:
                rm = (
                    self.momentum * mean.reshape(-1)
                    + (1 - self.momentum) * running_mean.data
                )
                rv = (
                    self.momentum * var.reshape(-1)
                    + (1 - self.momentum) * running_var.data
                )
                running_mean.data = rm
                running_var.data = rv

        else:
            mean = running_mean.data.reshape(1, -1, *([1] * (a.ndim - 2)))
            var = running_var.data.reshape(1, -1, *([1] * (a.ndim - 2)))

        rstd = 1.0 / lib_.sqrt(var + self.eps)
        xhat = (a.data - mean) * rstd

        out = xhat
        if self.has_weight:
            out = out * w.data.reshape(1, -1, *([1] * (a.ndim - 2)))
        if self.has_bias:
            out = out + b.data.reshape(1, -1, *([1] * (a.ndim - 2)))

        self._xhat = xhat
        self._rstd = rstd
        self._axes = axes
        self._m = m
        self._use_batch_stats = use_batch_stats

        self.result = Tensor(out, device=device)
        return self.result, functools.partial(self.__grad__, a=a, w=w, lib_=lib_)

    def __grad__(self, a: Tensor, w: Tensor, lib_: ModuleType) -> _GradType:
        if self.result is None or self.result.grad is None:
            raise RuntimeError("batch_norm backward called before forward.")

        if self._rstd is None or self._axes is None or self._m is None:
            raise RuntimeError("batch_norm cached data missing.")

        dy = _clone_array(self.result.grad)
        if dy.dtype != a.data.dtype:
            dy = dy.astype(a.data.dtype)

        C = a.shape[1]
        spatial_rank = a.ndim - 2
        view_shape = (1, C, *([1] * spatial_rank))
        perm = (1, 0, *range(2, a.ndim))

        if self.has_weight:
            gamma = _clone_array(w.data.reshape(view_shape))
            dy_gamma = dy * gamma
        else:
            dy_gamma = dy

        if self._use_batch_stats:
            xhat = _clone_array(self._xhat)
            rstd = _clone_array(self._rstd)
            dy_gamma_flat = dy_gamma.transpose(perm).reshape(C, -1)
            xhat_flat = xhat.transpose(perm).reshape(C, -1)

            mean_dy = dy_gamma_flat.mean(axis=1).reshape(view_shape)
            mean_dyxhat = (dy_gamma_flat * xhat_flat).mean(axis=1).reshape(view_shape)
            dx = (dy_gamma - mean_dy - xhat * mean_dyxhat) * rstd
        else:
            rstd = _clone_array(self._rstd)
            dx = dy_gamma * rstd

        dy_flat = dy.transpose(perm).reshape(C, -1)
        dweight = (
            (dy_flat * _clone_array(self._xhat).transpose(perm).reshape(C, -1)).sum(
                axis=1
            )
            if self.has_weight
            else None
        )
        dbias = dy_flat.sum(axis=1) if self.has_bias else None

        return dx, None, None, dweight, dbias


class group_norm_kernel(Operation):
    def __init__(
        self,
        num_groups: int,
        eps: float = 1e-5,
        has_weight: bool = True,
        has_bias: bool = True,
    ) -> None:
        super().__init__()
        self.num_groups = int(num_groups)
        self.eps = float(eps)
        self.has_weight = bool(has_weight)
        self.has_bias = bool(has_bias)

        self._xhat = None
        self._rstd = None
        self._group_shape = None
        self._reduce_axes = None
        self._m = None

    def clear(self) -> None:
        super().clear()
        self._xhat = None
        self._rstd = None
        self._group_shape = None
        self._reduce_axes = None
        self._m = None

    @func_op(n_in=3, n_ret=1, device="cpu")
    def cpu(self, a: Tensor, w: Tensor, b: Tensor) -> _FuncOpReturnType:
        return self._forward(a, w, b, lib_=np, device="cpu")

    @func_op(n_in=3, n_ret=1, device="gpu")
    def gpu(self, a: Tensor, w: Tensor, b: Tensor) -> _FuncOpReturnType:
        return self._forward(a, w, b, lib_=mx, device="gpu")

    def _forward(
        self,
        a: Tensor,
        w: Tensor,
        b: Tensor,
        lib_: ModuleType,
        device: _DeviceType,
    ) -> _FuncOpReturnType:
        N, C, *spatial = a.shape
        if C % self.num_groups != 0:
            raise ValueError("num_groups must divide channels.")

        group_size = C // self.num_groups
        group_shape = (N, self.num_groups, group_size, *spatial)
        x = a.data.reshape(group_shape)
        reduce_axes = (2,) + tuple(range(3, x.ndim))
        m = int(np.prod([x.shape[i] for i in reduce_axes]))

        mean = lib_.mean(x, axis=reduce_axes, keepdims=True)
        var = lib_.var(x, axis=reduce_axes, keepdims=True)
        rstd = 1.0 / lib_.sqrt(var + self.eps)
        xhat = (x - mean) * rstd

        out = xhat.reshape(a.shape)
        if self.has_weight:
            out = out * w.data.reshape(1, C, *([1] * len(spatial)))
        if self.has_bias:
            out = out + b.data.reshape(1, C, *([1] * len(spatial)))

        self._xhat = xhat
        self._rstd = rstd
        self._group_shape = group_shape
        self._reduce_axes = reduce_axes
        self._m = m

        self.result = Tensor(out, device=device)
        return self.result, functools.partial(self.__grad__, a=a, w=w, b=b, lib_=lib_)

    def __grad__(self, a: Tensor, w: Tensor, b: Tensor, lib_: ModuleType) -> _GradType:
        if self.result is None or self.result.grad is None:
            raise RuntimeError("group_norm backward called before forward.")

        if (
            self._xhat is None
            or self._rstd is None
            or self._group_shape is None
            or self._reduce_axes is None
            or self._m is None
        ):
            raise RuntimeError("group_norm cached data missing.")

        dy = self.result.grad
        N, C, *spatial = a.shape
        dy_g = dy.reshape(self._group_shape)
        xhat = self._xhat
        rstd = self._rstd
        axes = self._reduce_axes
        m = self._m

        if self.has_weight:
            w_broadcast = w.data.reshape(1, C, *([1] * len(spatial)))
            dyw = dy * w_broadcast
            dyw_g = dyw.reshape(self._group_shape)
        else:
            dyw_g = dy_g

        sum1 = lib_.sum(dyw_g, axis=axes, keepdims=True)
        sum2 = lib_.sum(dyw_g * xhat, axis=axes, keepdims=True)
        dx_g = (1.0 / m) * rstd * (m * dyw_g - sum1 - xhat * sum2)
        dx = dx_g.reshape(a.shape)

        reduce_axes = (0,) + tuple(range(2, a.ndim))
        dweight = lib_.sum(dy * xhat.reshape(a.shape), axis=reduce_axes)
        dbias = lib_.sum(dy, axis=reduce_axes)

        return dx, dweight, dbias



import lucid
import lucid.nn as nn

from lucid.types import _ShapeLike

from lucid._tensor import Tensor
from lucid.nn._kernel.norm import (
    layer_norm_kernel,
    batch_norm_kernel,
    group_norm_kernel,
)


def normalize(
    input_: Tensor, ord: int = 2, axis: int = 1, eps: float = 1e-12
) -> Tensor:
    norm = lucid.sum(lucid.abs(input_) ** ord, axis=axis, keepdims=True) ** (1 / ord)
    norm = lucid.maximum(norm, lucid.tensor(eps))

    return input_ / norm


def batch_norm(
    input_: Tensor,
    running_mean: Tensor | None,
    running_var: Tensor | None,
    weight: Tensor | None = None,
    bias: Tensor | None = None,
    training: bool = True,
    momentum: float = 0.1,
    eps: float = 1e-5,
) -> Tensor:
    C = input_.shape[1]
    running_mean_ = (
        running_mean
        if running_mean is not None
        else lucid.zeros((C,), device=input_.device, dtype=input_.dtype)
    )
    running_var_ = (
        running_var
        if running_var is not None
        else lucid.ones((C,), device=input_.device, dtype=input_.dtype)
    )
    weight_ = (
        weight
        if weight is not None
        else lucid.ones((C,), device=input_.device, dtype=input_.dtype)
    )
    bias_ = (
        bias
        if bias is not None
        else lucid.zeros((C,), device=input_.device, dtype=input_.dtype)
    )

    op = batch_norm_kernel(
        eps=eps,
        momentum=momentum,
        training=training,
        has_running=running_mean is not None and running_var is not None,
        has_weight=weight is not None,
        has_bias=bias is not None,
    )
    return op(input_, running_mean_, running_var_, weight_, bias_)


def layer_norm(
    input_: Tensor,
    normalized_shape: _ShapeLike,
    weight: Tensor | None = None,
    bias: Tensor | None = None,
    eps: float = 1e-5,
) -> Tensor:
    if input_.shape[-len(normalized_shape) :] != normalized_shape:
        raise ValueError(
            "Input tensor's normalized shape must match "
            + "the provided `normalized_shape`."
        )
    if weight is None:
        weight = lucid.ones(normalized_shape, device=input_.device)
        has_weight = False
    else:
        has_weight = True

    if bias is None:
        bias = lucid.zeros(normalized_shape, device=input_.device)
        has_bias = False
    else:
        has_bias = True

    op = layer_norm_kernel(
        normalized_shape, eps=eps, has_weight=has_weight, has_bias=has_bias
    )
    return op(input_, weight, bias)


def instance_norm(
    input_: Tensor,
    running_mean: Tensor | None,
    running_var: Tensor | None,
    weight: Tensor | None = None,
    bias: Tensor | None = None,
    training: bool = True,
    momentum: float = 0.1,
    eps: float = 1e-5,
) -> Tensor:
    C = input_.shape[1]
    spatial_dims = input_.shape[2:]
    use_instance_stats = training or running_mean is None or running_var is None
    axes = tuple(range(2, input_.ndim))

    if use_instance_stats:
        instance_mean = input_.mean(axis=axes, keepdims=True)
        instance_var = input_.var(axis=axes, keepdims=True)

        if training and running_mean is not None and running_var is not None:
            running_stats_ = (
                momentum * instance_mean.mean(axis=0).flatten()
                + (1 - momentum) * running_mean,
                momentum * instance_var.mean(axis=0).flatten()
                + (1 - momentum) * running_var,
            )

            if isinstance(running_mean, nn.Buffer) and isinstance(
                running_var, nn.Buffer
            ):
                running_mean.data = running_stats_[0].data
                running_var.data = running_stats_[1].data
            else:
                running_mean, running_var = running_stats_

        mean = instance_mean
        var = instance_var
    else:
        mean = running_mean.reshape(1, C, *(1,) * len(spatial_dims))
        var = running_var.reshape(1, C, *(1,) * len(spatial_dims))

    normalized = (input_ - mean) / lucid.sqrt(var + eps)

    if weight is not None:
        weight = weight.reshape((1, C) + (1,) * len(spatial_dims))
        normalized *= weight

    if bias is not None:
        bias = bias.reshape((1, C) + (1,) * len(spatial_dims))
        normalized += bias

    return normalized


def group_norm(
    input_: Tensor,
    num_groups: int,
    weight: Tensor | None,
    bias: Tensor | None,
    eps: float = 1e-5,
) -> Tensor:
    C = input_.shape[1]
    if weight is None:
        weight = lucid.ones((C,), device=input_.device)
        has_weight = False
    else:
        has_weight = True

    if bias is None:
        bias = lucid.zeros((C,), device=input_.device)
        has_bias = False
    else:
        has_bias = True

    op = group_norm_kernel(
        num_groups=num_groups, eps=eps, has_weight=has_weight, has_bias=has_bias
    )
    return op(input_, weight, bias)


def global_response_norm(
    input_: Tensor, gamma: Tensor, beta: Tensor, eps: float = 1e-6
) -> Tensor:
    Gx: Tensor = lucid.linalg.norm(input_, ord=2, axis=(-1, -2), keepdims=True)
    Nx = Gx / (Gx.mean(axis=-1, keepdims=True) + eps)
    return gamma * (input_ * Nx) + beta * input_



from dataclasses import dataclass

import lucid
from lucid._jit.ir import IRGraph, IRNode
from lucid._tensor.tensor import Tensor as _Tensor


class CompiledPlan:
    def __init__(self, graph: IRGraph, training: bool) -> None:
        self.graph = graph
        self.training = training
        self.exec_order: list[IRNode] = list(graph.nodes)

    def __repr__(self) -> str:
        return (
            f"CompiledPlan(nodes={len(self.exec_order)}, " f"training={self.training})"
        )


@dataclass
class ForwardResult:
    grad_func_map: dict
    value_map: dict
    exec_order: list
    leaf_vids: frozenset
    node_device_map: dict


class BackwardExecutor:
    def execute(
        self,
        forward_result: ForwardResult,
        output_vids: list,
        output_upstream_grads: dict,
        retain_grad: bool = False,
    ) -> None:
        grad_map: dict = {}
        for vid in output_vids:
            g = output_upstream_grads.get(vid)
            if g is None:
                t = forward_result.value_map.get(vid)
                if t is not None:
                    g = t.grad
            if g is not None:
                grad_map[vid] = g

        for node in reversed(forward_result.exec_order):
            if not node.has_gradient:
                continue

            device = forward_result.node_device_map[node.node_id]

            for out_id in node.output_ids:
                upstream = grad_map.get(out_id)
                if upstream is None:
                    continue
                grad_func = forward_result.grad_func_map.get(out_id)
                if grad_func is None:
                    continue

                result_tensor = forward_result.value_map[out_id]
                result_tensor.grad = upstream

                raw_grads = grad_func()
                if not isinstance(raw_grads, tuple):
                    raw_grads = (raw_grads,)

                for in_vid, g in zip(node.input_ids, raw_grads):
                    if g is None:
                        continue

                    in_tensor = forward_result.value_map.get(in_vid)
                    if in_tensor is None or not in_tensor.requires_grad:
                        continue

                    matched = lucid._match_grad_shape(in_tensor.data, g, device=device)
                    if in_vid in forward_result.leaf_vids:
                        lucid._set_tensor_grad(in_tensor, matched)
                    else:
                        existing = grad_map.get(in_vid)
                        grad_map[in_vid] = (
                            matched if existing is None else existing + matched
                        )

                if not retain_grad:
                    result_tensor.grad = None


class ForwardExecutor:
    def execute(
        self,
        plan: CompiledPlan,
        inputs: tuple,
        param_map: dict,
        grad_enabled: bool,
    ) -> tuple:
        graph = plan.graph
        value_map: dict = {}

        for vid, tensor in zip(graph.input_ids, inputs):
            value_map[vid] = tensor

        value_map.update(param_map)

        for vid in graph.constant_ids:
            iv = graph.values.get(vid)
            if iv is not None and iv.live_tensor is not None:
                value_map[vid] = iv.live_tensor

        requires_grad_any = grad_enabled and (
            any(getattr(t, "requires_grad", False) for t in inputs)
            or any(getattr(t, "requires_grad", False) for t in param_map.values())
        )

        build_backward = plan.training and requires_grad_any and grad_enabled
        grad_func_map: dict = {}

        for node in plan.exec_order:
            input_tensors = tuple(value_map[vid] for vid in node.input_ids)
            all_args = input_tensors + node.non_tensor_args

            op = node.op_class(*node.op_init_args, **node.op_init_kwargs)
            op_class = type(op)
            if node.device == "gpu":
                raw_func = getattr(op_class.gpu, "__wrapped__", None)
            else:
                raw_func = getattr(op_class.cpu, "__wrapped__", None)

            if raw_func is not None:
                raw_result = raw_func(op, *all_args, **node.non_tensor_kwargs)
            else:
                raw_result = op(
                    *input_tensors, *node.non_tensor_args, **node.non_tensor_kwargs
                )
                n_ret = node.n_ret
                if n_ret == 1 or not isinstance(raw_result, tuple):
                    raw_results = (raw_result,)
                else:
                    raw_results = raw_result
                for out_id, result_tensor in zip(node.output_ids, raw_results):
                    value_map[out_id] = result_tensor
                continue

            n_ret = node.n_ret
            if n_ret == 1:
                pairs = (raw_result,)
            elif n_ret is None:
                if (
                    isinstance(raw_result, tuple)
                    and len(raw_result) == 2
                    and isinstance(raw_result[0], _Tensor)
                ):
                    pairs = (raw_result,)
                else:
                    pairs = raw_result
            else:
                pairs = raw_result

            result_needs_grad = requires_grad_any and node.has_gradient and grad_enabled

            for out_id, (result_tensor, grad_func) in zip(node.output_ids, pairs):
                result_tensor.requires_grad = result_needs_grad

                if build_backward and result_needs_grad:
                    grad_func_map[out_id] = grad_func

                value_map[out_id] = result_tensor

        outputs = tuple(value_map[vid] for vid in graph.output_ids)

        fwd_result = None
        if build_backward:
            fwd_result = ForwardResult(
                grad_func_map=grad_func_map,
                value_map=value_map,
                exec_order=plan.exec_order,
                leaf_vids=frozenset(graph.input_ids) | graph.param_ids,
                node_device_map={n.node_id: n.device for n in plan.exec_order},
            )

        return outputs, fwd_result



from typing import Any, Callable, ClassVar

import numpy as np
import torch

import lucid.nn as nn

from lucid.test.core import ModuleTorchBase, ModuleTorchCase


def _unwrap_callable(operation: Any) -> Callable[..., Any]:
    if isinstance(operation, staticmethod):
        return operation.__func__

    if callable(operation):
        return operation

    raise TypeError(f"Expected callable operation, got: {type(operation)!r}")


class _NNModuleTorchBase(ModuleTorchBase):
    _collect_nested_test_classes = True

    case_name: ClassVar[str]
    inputs: ClassVar[tuple[Any, ...]]

    module_ctor: ClassVar[Callable[..., Any]]
    torch_module_ctor: ClassVar[Callable[..., Any] | None]
    expected_forward: ClassVar[Any | None] = None

    module_kwargs: ClassVar[dict[str, Any]] = {}
    torch_module_kwargs: ClassVar[dict[str, Any]] = {}

    requires_grad: ClassVar[bool | tuple[bool, ...]] = True
    expected_input_grads: ClassVar[tuple[object | None, ...] | None] = None
    module_train_modes: ClassVar[tuple[bool, ...]] = (True,)

    compare_grad_with_torch: ClassVar[bool] = True
    compare_state_dict: ClassVar[bool] = True

    check_finite: ClassVar[bool] = True
    rtol: ClassVar[float] = 1e-7
    atol: ClassVar[float] = 1e-8

    def tensor_op_devices(self) -> tuple[str, ...]:
        return ("cpu", "gpu")

    def _normalize_bool_flags(
        self, flags: bool | tuple[bool, ...], expected_len: int
    ) -> tuple[bool, ...]:
        if isinstance(flags, bool):
            return (bool(flags),) * expected_len
        if len(flags) == 1 and expected_len > 1:
            return (bool(flags[0]),) * expected_len
        if len(flags) != expected_len:
            raise AssertionError(
                f"requires_grad length {len(flags)} does not match inputs "
                f"(expected {expected_len}) for case '{self.case_name}'."
            )
        return tuple(bool(v) for v in flags)

    def _build_inputs(self) -> tuple[Any, ...]:
        requires_grad = self._normalize_bool_flags(self.requires_grad, len(self.inputs))
        outputs: list[Any] = []

        for item, req_grad in zip(self.inputs, requires_grad):
            outputs.append(self.tensor(item, requires_grad=req_grad))

        return tuple(outputs)

    def _build_case(self) -> ModuleTorchCase:
        module_ctor = _unwrap_callable(self.module_ctor)
        torch_module_ctor = _unwrap_callable(self.torch_module_ctor)

        requires_grad = self._normalize_bool_flags(self.requires_grad, len(self.inputs))

        return ModuleTorchCase(
            name=self.case_name,
            build_inputs=lambda _: self._build_inputs(),
            module_ctor=module_ctor,
            torch_module_ctor=torch_module_ctor,
            module_kwargs=self.module_kwargs,
            torch_module_kwargs=self.torch_module_kwargs,
            expected_forward=self.expected_forward,
            expected_input_grads=self.expected_input_grads,
            requires_grad=requires_grad,
            module_train_modes=self.module_train_modes,
            compare_grad_with_torch=self.compare_grad_with_torch,
            compare_state_dict=self.compare_state_dict,
            check_finite=self.check_finite,
            rtol=self.rtol,
            atol=self.atol,
        )

    def module_op_cases(self) -> tuple[ModuleTorchCase, ...]:
        return (self._build_case(),)


class TestIdentity(_NNModuleTorchBase):
    __test__ = False

    class TestDefault(_NNModuleTorchBase):
        case_name = "identity"
        inputs = (np.array([[1.0, -2.0], [3.0, 4.0]], dtype=np.float64),)
        module_ctor = nn.Identity
        torch_module_ctor = torch.nn.Identity
        expected_forward = np.array([[1.0, -2.0], [3.0, 4.0]], dtype=np.float64)
        compare_state_dict = False


class TestFlatten(_NNModuleTorchBase):
    __test__ = False

    class TestDefault(_NNModuleTorchBase):
        case_name = "flatten_default"
        inputs = (np.array([[1.0, 2.0], [3.0, 4.0]], dtype=np.float64),)
        module_ctor = nn.Flatten
        torch_module_ctor = torch.nn.Flatten
        module_kwargs = {}
        torch_module_kwargs = {}
        expected_forward = np.array([1.0, 2.0, 3.0, 4.0], dtype=np.float64)

    class TestStartAxis(_NNModuleTorchBase):
        case_name = "flatten_start_axis_end_axis"
        inputs = (np.arange(12.0, dtype=np.float64).reshape(2, 2, 3),)
        module_ctor = nn.Flatten
        torch_module_ctor = torch.nn.Flatten
        module_kwargs = {"start_axis": 0, "end_axis": -1}
        torch_module_kwargs = {"start_dim": 0, "end_dim": -1}
        expected_forward = np.arange(12.0, dtype=np.float64).reshape(2, 2 * 3)


class TestActivation(_NNModuleTorchBase):
    __test__ = False

    class TestReLU(_NNModuleTorchBase):
        case_name = "relu"
        inputs = (np.array([[-1.0, 2.0], [3.0, -4.0]], dtype=np.float64),)
        module_ctor = nn.ReLU
        torch_module_ctor = torch.nn.ReLU

    class TestLeakyReLU(_NNModuleTorchBase):
        case_name = "leaky_relu"
        inputs = (np.array([[-1.0, 2.0], [3.0, -4.0]], dtype=np.float64),)
        module_ctor = nn.LeakyReLU
        torch_module_ctor = torch.nn.LeakyReLU
        module_kwargs = {"negative_slope": 0.2}
        torch_module_kwargs = {"negative_slope": 0.2}

    class TestTanh(_NNModuleTorchBase):
        case_name = "tanh"
        inputs = (np.array([[-1.0, 0.5], [3.0, -4.0]], dtype=np.float64),)
        module_ctor = nn.Tanh
        torch_module_ctor = torch.nn.Tanh


class TestLinear(_NNModuleTorchBase):
    __test__ = False

    class TestDefault(_NNModuleTorchBase):
        case_name = "linear"
        inputs = (np.array([[1.0, 2.0], [3.0, 4.0]], dtype=np.float64),)
        module_ctor = nn.Linear
        torch_module_ctor = torch.nn.Linear
        module_kwargs = {"in_features": 2, "out_features": 3, "bias": True}
        torch_module_kwargs = {"in_features": 2, "out_features": 3, "bias": True}

    class TestNoBias(_NNModuleTorchBase):
        case_name = "linear_no_bias"
        inputs = (np.array([[1.0, 2.0], [3.0, 4.0]], dtype=np.float64),)
        module_ctor = nn.Linear
        torch_module_ctor = torch.nn.Linear
        module_kwargs = {"in_features": 2, "out_features": 3, "bias": False}
        torch_module_kwargs = {"in_features": 2, "out_features": 3, "bias": False}


class TestConv2d(_NNModuleTorchBase):
    case_name = "conv2d"
    inputs = (np.arange(1.0 * 1 * 4 * 4, dtype=np.float64).reshape(1, 1, 4, 4),)
    module_ctor = nn.Conv2d
    torch_module_ctor = torch.nn.Conv2d
    module_kwargs = {
        "in_channels": 1,
        "out_channels": 2,
        "kernel_size": 3,
        "padding": 1,
    }
    torch_module_kwargs = {
        "in_channels": 1,
        "out_channels": 2,
        "kernel_size": 3,
        "padding": 1,
    }
    rtol = 1e-5
    atol = 1e-6


class TestBatchNorm(_NNModuleTorchBase):
    __test__ = False

    class TestBatchNorm1dTrain(_NNModuleTorchBase):
        case_name = "batch_norm1d_train"
        inputs = (np.array([[[1.0, 2.0, 3.0]]], dtype=np.float64),)
        module_ctor = nn.BatchNorm1d
        torch_module_ctor = torch.nn.BatchNorm1d
        module_kwargs = {"num_features": 1, "momentum": 0.1}
        torch_module_kwargs = {"num_features": 1, "momentum": 0.1}
        module_train_modes = (True,)

    class TestBatchNorm1dEval(_NNModuleTorchBase):
        case_name = "batch_norm1d_eval"
        inputs = (np.array([[[1.0, 2.0, 3.0]]], dtype=np.float64),)
        module_ctor = nn.BatchNorm1d
        torch_module_ctor = torch.nn.BatchNorm1d
        module_kwargs = {"num_features": 1, "momentum": 0.1}
        torch_module_kwargs = {"num_features": 1, "momentum": 0.1}
        module_train_modes = (False,)

    class TestBatchNorm2dTrain(_NNModuleTorchBase):
        case_name = "batch_norm2d_train"
        inputs = (
            np.array(
                [
                    [
                        [[1.0, 2.0], [3.0, 4.0]],
                        [[5.0, 6.0], [7.0, 8.0]],
                    ]
                ],
                dtype=np.float64,
            ),
        )
        module_ctor = nn.BatchNorm2d
        torch_module_ctor = torch.nn.BatchNorm2d
        module_kwargs = {"num_features": 2, "momentum": 0.1}
        torch_module_kwargs = {"num_features": 2, "momentum": 0.1}
        module_train_modes = (True,)

    class TestBatchNorm2dEval(_NNModuleTorchBase):
        case_name = "batch_norm2d_eval"
        inputs = (
            np.array(
                [
                    [
                        [[1.0, 2.0], [3.0, 4.0]],
                        [[5.0, 6.0], [7.0, 8.0]],
                    ]
                ],
                dtype=np.float64,
            ),
        )
        module_ctor = nn.BatchNorm2d
        torch_module_ctor = torch.nn.BatchNorm2d
        module_kwargs = {"num_features": 2, "momentum": 0.1}
        torch_module_kwargs = {"num_features": 2, "momentum": 0.1}
        module_train_modes = (False,)


class TestLayerNorm(_NNModuleTorchBase):
    case_name = "layer_norm"
    inputs = (np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]], dtype=np.float64),)
    module_ctor = nn.LayerNorm
    torch_module_ctor = torch.nn.LayerNorm
    module_kwargs = {"normalized_shape": (3,)}
    torch_module_kwargs = {"normalized_shape": (3,)}
    rtol = 1e-5
    atol = 1e-6


class TestGroupNorm(_NNModuleTorchBase):
    case_name = "group_norm"
    inputs = (np.arange(1.0 * 2 * 2 * 2, dtype=np.float64).reshape(1, 2, 2, 2),)
    module_ctor = nn.GroupNorm
    torch_module_ctor = torch.nn.GroupNorm
    module_kwargs = {"num_groups": 1, "num_channels": 2}
    torch_module_kwargs = {"num_groups": 1, "num_channels": 2}


class TestEmbedding(_NNModuleTorchBase):
    case_name = "embedding"
    inputs = (np.array([[1, 3, 4], [2, 0, 5]], dtype=np.int64),)
    module_ctor = nn.Embedding
    torch_module_ctor = torch.nn.Embedding
    module_kwargs = {"num_embeddings": 10, "embedding_dim": 4}
    torch_module_kwargs = {"num_embeddings": 10, "embedding_dim": 4}
    requires_grad = (False,)
    module_train_modes = (False,)

    def tensor_op_devices(self) -> tuple[str, ...]:
        return ("cpu",)


class TestDropout(_NNModuleTorchBase):
    case_name = "dropout_eval"
    inputs = (np.array([[-1.0, 0.0, 1.0], [2.0, -3.0, 4.0]], dtype=np.float64),)
    module_ctor = nn.Dropout
    torch_module_ctor = torch.nn.Dropout
    module_kwargs = {"p": 0.2}
    torch_module_kwargs = {"p": 0.2}
    module_train_modes = (False,)


class TestPooling(_NNModuleTorchBase):
    __test__ = False

    class TestAvgPool2d(_NNModuleTorchBase):
        case_name = "avg_pool2d"
        inputs = (
            np.array(
                [
                    [
                        [
                            [1.0, 2.0, 3.0, 4.0],
                            [5.0, 6.0, 7.0, 8.0],
                            [9.0, 10.0, 11.0, 12.0],
                            [13.0, 14.0, 15.0, 16.0],
                        ]
                    ]
                ],
                dtype=np.float64,
            ),
        )
        module_ctor = nn.AvgPool2d
        torch_module_ctor = torch.nn.AvgPool2d
        module_kwargs = {"kernel_size": 2, "stride": 2}
        torch_module_kwargs = {"kernel_size": 2, "stride": 2}

    class TestMaxPool2d(_NNModuleTorchBase):
        case_name = "max_pool2d"
        inputs = (
            np.array(
                [
                    [
                        [
                            [1.0, 2.0, 3.0, 4.0],
                            [5.0, 6.0, 7.0, 8.0],
                            [9.0, 10.0, 11.0, 12.0],
                            [13.0, 14.0, 15.0, 16.0],
                        ]
                    ]
                ],
                dtype=np.float64,
            ),
        )
        module_ctor = nn.MaxPool2d
        torch_module_ctor = torch.nn.MaxPool2d
        module_kwargs = {"kernel_size": 2, "stride": 2}
        torch_module_kwargs = {"kernel_size": 2, "stride": 2}



from dataclasses import dataclass
import math

import lucid
import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor

__all__ = ["Transformer", "TransformerConfig", "transformer_base", "transformer_big"]


@dataclass
class TransformerConfig:
    src_vocab_size: int
    tgt_vocab_size: int
    d_model: int
    num_heads: int
    num_encoder_layers: int
    num_decoder_layers: int
    dim_feedforward: int
    dropout: float = 0.1
    max_len: int = 5000

    def __post_init__(self) -> None:
        if self.src_vocab_size <= 0:
            raise ValueError("src_vocab_size must be greater than 0")
        if self.tgt_vocab_size <= 0:
            raise ValueError("tgt_vocab_size must be greater than 0")
        if self.d_model <= 0:
            raise ValueError("d_model must be greater than 0")
        if self.num_heads <= 0:
            raise ValueError("num_heads must be greater than 0")
        if self.d_model % self.num_heads != 0:
            raise ValueError("d_model must be divisible by num_heads")
        if self.num_encoder_layers <= 0:
            raise ValueError("num_encoder_layers must be greater than 0")
        if self.num_decoder_layers <= 0:
            raise ValueError("num_decoder_layers must be greater than 0")
        if self.dim_feedforward <= 0:
            raise ValueError("dim_feedforward must be greater than 0")
        if self.dropout < 0 or self.dropout >= 1:
            raise ValueError("dropout must be in the range [0, 1)")
        if self.max_len <= 0:
            raise ValueError("max_len must be greater than 0")


class _PositionalEncoding(nn.Module):
    def __init__(self, d_model: int, dropout: float = 0.1, max_len: int = 5000) -> None:
        super().__init__()
        self.dropout = nn.Dropout(dropout)

        pe = lucid.zeros(max_len, d_model)
        position = lucid.arange(0, max_len).unsqueeze(axis=1)
        div_term = lucid.exp(lucid.arange(0, d_model, 2) * (-math.log(1e4) / d_model))

        pe[:, 0::2] = lucid.sin(position * div_term)
        pe[:, 1::2] = lucid.cos(position * div_term)

        pe = pe.unsqueeze(axis=1)
        self.register_buffer("pe", pe)

    def forward(self, x: Tensor) -> Tensor:
        x += self.pe[: x.shape[0]]
        x = self.dropout(x)

        return x


class Transformer(nn.Module):
    def __init__(self, config: TransformerConfig) -> None:
        super().__init__()
        self.config = config
        self.d_model = config.d_model

        self.src_tok_emb = nn.Embedding(config.src_vocab_size, config.d_model)
        self.tgt_tok_emb = nn.Embedding(config.tgt_vocab_size, config.d_model)

        self.positional_encoding = _PositionalEncoding(
            config.d_model, config.dropout, config.max_len
        )

        self.transformer = nn.Transformer(
            d_model=config.d_model,
            num_heads=config.num_heads,
            num_encoder_layers=config.num_encoder_layers,
            num_decoder_layers=config.num_decoder_layers,
            dim_feedforward=config.dim_feedforward,
            dropout=config.dropout,
        )
        self.fc_out = nn.Linear(config.d_model, config.tgt_vocab_size)

    def forward(
        self,
        src: Tensor,
        tgt: Tensor,
        src_mask: Tensor | None = None,
        tgt_mask: Tensor | None = None,
        mem_mask: Tensor | None = None,
        src_key_padding_mask: Tensor | None = None,
        tgt_key_padding_mask: Tensor | None = None,
        mem_key_padding_mask: Tensor | None = None,
        kv_cache: nn.KVCache | None = None,
        use_cache: bool = False,
        cache_position: Tensor | None = None,
        cache_start_layer_idx: int = 0,
        encoder_kv_cache: nn.KVCache | None = None,
        use_encoder_cache: bool = False,
        encoder_cache_position: Tensor | None = None,
        encoder_cache_start_layer_idx: int = 0,
    ) -> Tensor:
        src_emb = self.src_tok_emb(src) * math.sqrt(self.d_model)
        tgt_emb = self.tgt_tok_emb(tgt) * math.sqrt(self.d_model)

        src_emb = self.positional_encoding(src_emb)
        tgt_emb = self.positional_encoding(tgt_emb)

        output = self.transformer(
            src_emb,
            tgt_emb,
            src_mask=src_mask,
            tgt_mask=tgt_mask,
            mem_mask=mem_mask,
            src_key_padding_mask=src_key_padding_mask,
            tgt_key_padding_mask=tgt_key_padding_mask,
            mem_key_padding_mask=mem_key_padding_mask,
            kv_cache=kv_cache,
            use_cache=use_cache,
            cache_position=cache_position,
            cache_start_layer_idx=cache_start_layer_idx,
            encoder_kv_cache=encoder_kv_cache,
            use_encoder_cache=use_encoder_cache,
            encoder_cache_position=encoder_cache_position,
            encoder_cache_start_layer_idx=encoder_cache_start_layer_idx,
        )
        output = self.fc_out(output)

        return output


@register_model
def transformer_base(
    src_vocab_size: int = 12000, tgt_vocab_size: int = 12000, **kwargs
) -> Transformer:
    config_kwargs = {
        "src_vocab_size": src_vocab_size,
        "tgt_vocab_size": tgt_vocab_size,
        "d_model": 512,
        "num_heads": 8,
        "num_encoder_layers": 6,
        "num_decoder_layers": 6,
        "dim_feedforward": 2048,
        "dropout": 0.1,
    }
    config_kwargs.update(kwargs)
    return Transformer(TransformerConfig(**config_kwargs))


@register_model
def transformer_big(
    src_vocab_size: int = 12000, tgt_vocab_size: int = 12000, **kwargs
) -> Transformer:
    config_kwargs = {
        "src_vocab_size": src_vocab_size,
        "tgt_vocab_size": tgt_vocab_size,
        "d_model": 1024,
        "num_heads": 16,
        "num_encoder_layers": 6,
        "num_decoder_layers": 6,
        "dim_feedforward": 4096,
        "dropout": 0.3,
    }
    config_kwargs.update(kwargs)
    return Transformer(TransformerConfig(**config_kwargs))



import json

from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import Iterable, TypeVar

import lucid
from lucid.types import _DeviceType

__all__ = ["SpecialTokens", "Tokenizer"]


class SpecialTokens(Enum):
    UNK: str = "[UNK]"
    PAD: str = "[PAD]"
    BOS: str = "[BOS]"
    EOS: str = "[EOS]"


_SpecialTokensOrStr = SpecialTokens | str

T = TypeVar("T", bound="Tokenizer")


class Tokenizer(ABC):
    def __init__(
        self,
        unk_token: _SpecialTokensOrStr = SpecialTokens.UNK,
        pad_token: _SpecialTokensOrStr = SpecialTokens.PAD,
        bos_token: _SpecialTokensOrStr | None = SpecialTokens.BOS,
        eos_token: _SpecialTokensOrStr | None = SpecialTokens.EOS,
    ) -> None:
        super().__init__()
        self.unk_token = self._resolve_token(unk_token)
        self.pad_token = self._resolve_token(pad_token)
        self.bos_token = self._resolve_token(bos_token)
        self.eos_token = self._resolve_token(eos_token)

    @staticmethod
    def _resolve_token(token: _SpecialTokensOrStr | None) -> str | None:
        if isinstance(token, SpecialTokens):
            return token.value
        return token

    @property
    @abstractmethod
    def vocab_size(self) -> int: ...

    @abstractmethod
    def tokenize(self, text: str) -> list[str]: ...

    @abstractmethod
    def convert_tokens_to_ids(self, tokens: str | list[str]) -> int | list[int]: ...

    @abstractmethod
    def convert_ids_to_tokens(self, ids: int | list[int]) -> str | list[str]: ...

    @abstractmethod
    def convert_tokens_to_string(self, tokens: list[str]) -> str: ...

    @property
    def all_special_tokens(self) -> list[str]:
        tokens = [self.unk_token, self.pad_token, self.bos_token, self.eos_token]
        return [t for t in tokens if t is not None]

    @property
    def all_special_ids(self) -> list[int]:
        ids = self.convert_tokens_to_ids(self.all_special_tokens)
        if not isinstance(ids, list):
            raise TypeError("convert_tokens_to_ids(list[str]) must return 'list[int]'.")
        return ids

    def build_inputs_with_special_tokens(self, tokens: list[str]) -> list[str]:
        out = tokens
        if self.bos_token is not None:
            out = [self.bos_token] + out
        if self.eos_token is not None:
            out = out + [self.eos_token]
        return out

    def encode(
        self,
        text: str,
        add_special_tokens: bool = True,
        return_tensor: bool = False,
        device: _DeviceType = "cpu",
    ) -> list[int] | lucid.LongTensor:
        tokens = self.tokenize(text)
        if add_special_tokens:
            tokens = self.build_inputs_with_special_tokens(tokens)

        ids = self.convert_tokens_to_ids(tokens)
        if not isinstance(ids, list):
            raise TypeError("convert_tokens_to_ids(list[str]) must return 'list[int]'.")

        if return_tensor:
            return lucid.LongTensor(ids, device=device)
        return ids

    def decode(
        self, ids: list[int] | lucid.LongTensor, skip_special_tokens: bool = True
    ) -> str:
        if isinstance(ids, lucid.Tensor):
            if ids.dtype.base_dtype is not int:
                raise TypeError(
                    "'ids' must be a 'lucid.LongTensor'. Try 'Tensor.long()'"
                )
            if ids.ndim != 1:
                raise ValueError(f"'ids' must be 1-D tensor, got {ids.ndim}-D tensor")

            ids = ids.tolist()

        tokens = self.convert_ids_to_tokens(ids)
        if not isinstance(tokens, list):
            raise TypeError("convert_ids_to_tokens(list[int]) must return 'list[str]'.")

        if skip_special_tokens:
            special = set(self.all_special_tokens)
            tokens = [t for t in tokens if t not in special]

        return self.convert_tokens_to_string(tokens)

    def batch_encode(
        self,
        texts: Iterable[str],
        add_special_tokens: bool = True,
        padding: bool = False,
        max_length: int | None = None,
        truncation: bool = False,
        return_tensor: bool = False,
        device: _DeviceType = "cpu",
    ) -> list[list[int]] | lucid.LongTensor:
        batch_ids = [
            self.encode(t, add_special_tokens=add_special_tokens) for t in texts
        ]
        if truncation and max_length is not None:
            batch_ids: list[list[int]] = [ids[:max_length] for ids in batch_ids]

        if return_tensor:
            if not padding:
                lens = {len(ids) for ids in batch_ids}
                if len(lens) != 1:
                    raise ValueError(
                        "Cannot return 2D tensor for ragged batch. "
                        "Set 'padding=True' (or pad manually)."
                    )

            if padding:
                if self.pad_token is None:
                    raise ValueError("'pad_token' is required when 'padding=True'")

                pad_id = self.convert_ids_to_tokens(self.pad_token)
                if isinstance(pad_id, list):
                    raise TypeError("convert_tokens_to_ids(str) must return int.")

                target_len = (
                    max_length
                    if max_length is not None
                    else max(len(ids) for ids in batch_ids)
                )
                padded: list[list[int]] = []
                for ids in batch_ids:
                    cur = ids[:target_len]
                    cur = cur + [pad_id] * max(0, target_len - len(cur))
                    padded.append(cur)

                batch_ids = padded

            return lucid.LongTensor(batch_ids, device=device)

        return batch_ids

    def batch_decode(
        self,
        batch_ids: Iterable[list[int]] | lucid.LongTensor,
        skip_special_tokens: bool = True,
    ) -> list[str]:
        return [
            self.decode(ids, skip_special_tokens=skip_special_tokens)
            for ids in batch_ids
        ]

    def __call__(
        self,
        text: str,
        add_special_tokens: bool = True,
        return_tensor: bool = False,
        device: _DeviceType = "cpu",
    ) -> list[int] | lucid.LongTensor:
        return self.encode(text, add_special_tokens, return_tensor, device)

    def fit(self, texts: Iterable[str], **kwargs) -> Tokenizer:
        raise NotImplementedError(
            f"{type(self).__name__} does not support vocabulary training via 'fit'."
        )

    @classmethod
    def train_from_iterator(cls, texts: Iterable[str], **kwargs) -> Tokenizer:
        raise NotImplementedError(
            f"{cls.__name__} does not support vocabulary training via "
            "'train_from_iterator'."
        )

    @staticmethod
    def _save_json(data: dict, path: Path | str) -> str:
        out_path = Path(path)
        with out_path.open("w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=2)
        return str(out_path)

    @staticmethod
    def _load_json(path: Path | str) -> dict[str, object]:
        with Path(path).open("r", encoding="utf-8") as f:
            data = json.load(f)
        if not isinstance(data, dict):
            raise TypeError(f"Expected JSON object at '{path}'.")
        return data

    def _save_tokenizer_config(self, save_directory: Path | str, config: dict) -> str:
        save_path = Path(save_directory)
        save_path.mkdir(parents=True, exist_ok=True)
        return self._save_json(config, save_path / "tokenizer_config.json")

    @classmethod
    def _load_tokenizer_config(
        cls,
        pretrained_model_name_or_path: Path | str,
        allowed_keys: set[str] | None = None,
    ) -> dict[str, object]:
        path = Path(pretrained_model_name_or_path)
        config_file = (
            path / "tokenizer_config.json"
            if path.is_dir()
            else path.with_name("tokenizer_config.json")
        )
        if not config_file.exists():
            return {}
        config = cls._load_json(config_file)
        if allowed_keys is None:
            return config
        return {k: v for k, v in config.items() if k in allowed_keys}

    def _build_id_to_token_map(
        self,
        vocab: dict[str, int],
        unk_token: str = SpecialTokens.UNK.value,
        verbose: bool = False,
    ) -> list[str]:
        from tqdm.auto import tqdm

        max_id = max(vocab.values(), default=-1)
        id_to_token = ["" for _ in range(max_id + 1)]

        items = vocab.items()
        if verbose:
            items = tqdm(
                items,
                total=len(vocab),
                desc=f"{type(self).__name__} id->token map",
                dynamic_ncols=True,
                unit="tok",
            )
        for token, idx in items:
            if idx < 0:
                raise ValueError("Vocabulary indices must be non-negative integers.")
            if idx >= len(id_to_token):
                id_to_token.extend([""] * (idx - len(id_to_token) + 1))
            if id_to_token[idx]:
                raise ValueError(f"Duplicate token index detected: {idx}")
            id_to_token[idx] = token

        for i, token in enumerate(id_to_token):
            if not token:
                id_to_token[i] = unk_token
        return id_to_token

    @abstractmethod
    def save_pretrained(self, save_directory: Path | str) -> list[str]: ...

    @classmethod
    @abstractmethod
    def from_pretrained(
        cls, pretrained_model_name_or_path: str | Path, **kwrags
    ) -> Tokenizer: ...



import functools
from types import ModuleType

import numpy as np

from lucid._backend.core import Operation, func_op, _FuncOpReturnType, _GradType
from lucid._backend.metal import mx
from lucid._tensor import Tensor
from lucid.types import _DeviceType, _TensorData


def _to_int(arr: _TensorData, lib_: ModuleType) -> _TensorData:
    if lib_ is np:
        return arr.astype(np.int64)
    return arr.astype(mx.int32)


class cross_entropy_kernel(Operation):
    def __init__(
        self,
        reduction: str | None = "mean",
        eps: float = 1e-7,
        ignore_index: int | None = None,
        has_weight: bool = True,
    ) -> None:
        super().__init__()
        self.reduction = reduction
        self.eps = float(eps)
        self.ignore_index = ignore_index
        self.has_weight = bool(has_weight)

        self._log_probs = None
        self._probs = None
        self._target = None
        self._weight = None
        self._valid_count = None

    def clear(self) -> None:
        super().clear()
        self._log_probs = None
        self._probs = None
        self._target = None
        self._weight = None
        self._valid_count = None

    @func_op(n_in=3, n_ret=1)
    def cpu(self, logits: Tensor, target: Tensor, weight: Tensor) -> _FuncOpReturnType:
        return self._forward(logits, target, weight, lib_=np, device="cpu")

    @func_op(n_in=3, n_ret=1, device="gpu")
    def gpu(self, logits: Tensor, target: Tensor, weight: Tensor) -> _FuncOpReturnType:
        return self._forward(logits, target, weight, lib_=mx, device="gpu")

    def _forward(
        self,
        logits: Tensor,
        target: Tensor,
        weight: Tensor,
        lib_: ModuleType,
        device: _DeviceType,
    ) -> _FuncOpReturnType:
        if logits.ndim != 2:
            raise ValueError("cross_entropy expects 2D logits [N, C].")

        N, _ = logits.shape
        max_val = lib_.max(logits.data, axis=1, keepdims=True)
        exp_x = lib_.exp(logits.data - max_val)
        sum_exp = lib_.sum(exp_x, axis=1, keepdims=True)
        log_probs = (logits.data - max_val) - lib_.log(sum_exp)
        probs = exp_x / sum_exp

        target_int = _to_int(target.data, lib_)
        if lib_ is np:
            idx = np.arange(N, dtype=np.int64)
        else:
            idx = mx.arange(N, dtype=mx.int32)

        if self.ignore_index is not None:
            mask = target_int != self.ignore_index
        else:
            mask = None

        gather = log_probs[idx, target_int]
        loss = -gather

        if self.has_weight:
            w = weight.data
            loss = loss * w[target_int]
        else:
            w = None

        if mask is not None:
            if lib_ is np:
                loss = loss * mask.astype(loss.dtype)
            else:
                loss = loss * mask.astype(loss.dtype)

        if self.reduction is None:
            out = loss
            valid_count = None
        elif self.reduction == "sum":
            out = lib_.sum(loss)
            valid_count = None
        elif self.reduction == "mean":
            if mask is None:
                valid_count = N
            else:
                valid_count = lib_.sum(mask)
                if hasattr(valid_count, "item") and valid_count.item() == 0:
                    out = lib_.zeros((), dtype=loss.dtype)
                    self.result = Tensor(out, device=device)
                    return self.result, functools.partial(self.__grad__, lib_=lib_)
            out = lib_.sum(loss) / valid_count
        else:
            raise ValueError("Invalid reduction type. Choose 'mean', 'sum', or 'none'.")

        self._log_probs = log_probs
        self._probs = probs
        self._target = target_int
        self._weight = w
        self._valid_count = valid_count

        self.result = Tensor(out, device=device)
        return self.result, functools.partial(self.__grad__, lib_=lib_)

    def __grad__(self, lib_: ModuleType) -> _GradType:
        if self.result is None or self.result.grad is None:
            raise RuntimeError("cross_entropy backward called before forward.")
        if self._probs is None or self._target is None:
            raise RuntimeError("cross_entropy cached data missing.")

        probs = self._probs
        target = self._target
        N = probs.shape[0]
        C = probs.shape[1]

        if lib_ is np:
            grad_input = probs.copy()
            idx = np.arange(N, dtype=np.int64)
        else:
            grad_input = mx.array(probs)
            idx = mx.arange(N, dtype=mx.int32)

        grad_input[idx, target] = grad_input[idx, target] - 1

        if self._weight is not None:
            grad_input = grad_input * self._weight[target][:, None]

        if self.ignore_index is not None:
            if lib_ is np:
                mask = (target != self.ignore_index).astype(grad_input.dtype)
                grad_input = grad_input * mask[:, None]
            else:
                mask = (target != self.ignore_index).astype(grad_input.dtype)
                grad_input = grad_input * mask[:, None]

        if self.reduction is None:
            go = self.result.grad
            grad_input = grad_input * go[:, None]
            dweight = None
        else:
            go = self.result.grad
            if self.reduction == "mean":
                if self._valid_count is None:
                    grad_input = grad_input / N
                else:
                    grad_input = grad_input / self._valid_count
            grad_input = grad_input * go

            dweight = None
            if self._weight is not None:
                if lib_ is np:
                    dweight = np.zeros((C,), dtype=grad_input.dtype)
                else:
                    dweight = mx.zeros((C,), dtype=grad_input.dtype)

                if self.reduction is None:
                    go_vec = go
                else:
                    go_vec = None

                for c in range(C):
                    if lib_ is np:
                        mask_c = target == c
                        if self.ignore_index is not None:
                            mask_c = mask_c & (target != self.ignore_index)
                        if go_vec is None:
                            contrib = -self._log_probs[mask_c, c]
                            if self.reduction == "mean":
                                denom = (
                                    self._valid_count
                                    if self._valid_count is not None
                                    else N
                                )
                                contrib = contrib / denom
                            dweight[c] = np.sum(contrib) * go
                        else:
                            contrib = -self._log_probs[mask_c, c]
                            dweight[c] = np.sum(contrib * go_vec[mask_c])
                    else:
                        mask_c = target == c
                        if self.ignore_index is not None:
                            mask_c = mask_c & (target != self.ignore_index)

                        contrib = -self._log_probs[mask_c, c]
                        if self.reduction == "mean":
                            denom = (
                                self._valid_count
                                if self._valid_count is not None
                                else N
                            )
                            contrib = contrib / denom

                        if go_vec is None:
                            dweight = dweight.at[c].add(mx.sum(contrib) * go)
                        else:
                            dweight = dweight.at[c].add(
                                mx.sum(contrib * go_vec[mask_c])
                            )

        return grad_input, None, dweight


class binary_cross_entropy_kernel(Operation):
    def __init__(
        self,
        reduction: str | None = "mean",
        eps: float = 1e-7,
        has_weight: bool = True,
    ) -> None:
        super().__init__()
        self.reduction = reduction
        self.eps = float(eps)
        self.has_weight = bool(has_weight)

        self._input = None
        self._target = None
        self._weight = None

    def clear(self) -> None:
        super().clear()
        self._input = None
        self._target = None
        self._weight = None

    @func_op(n_in=3, n_ret=1)
    def cpu(self, input_: Tensor, target: Tensor, weight: Tensor) -> _FuncOpReturnType:
        return self._forward(input_, target, weight, lib_=np, device="cpu")

    @func_op(n_in=3, n_ret=1, device="gpu")
    def gpu(self, input_: Tensor, target: Tensor, weight: Tensor) -> _FuncOpReturnType:
        return self._forward(input_, target, weight, lib_=mx, device="gpu")

    def _forward(
        self,
        input_: Tensor,
        target: Tensor,
        weight: Tensor,
        lib_: ModuleType,
        device: _DeviceType,
    ) -> _FuncOpReturnType:
        x = input_.data
        t = target.data
        x = lib_.clip(x, self.eps, 1.0 - self.eps)
        loss = -(t * lib_.log(x) + (1 - t) * lib_.log(1 - x))

        if self.has_weight:
            loss = loss * weight.data

        if self.reduction is None:
            out = loss
        elif self.reduction == "sum":
            out = lib_.sum(loss)
        elif self.reduction == "mean":
            out = lib_.mean(loss)
        else:
            raise ValueError("Invalid reduction type. Choose 'mean', 'sum', or 'none'.")

        self._input = x
        self._target = t
        self._weight = weight.data if self.has_weight else None

        self.result = Tensor(out, device=device)
        return self.result, functools.partial(self.__grad__)

    def __grad__(self) -> _GradType:
        if self.result is None or self.result.grad is None:
            raise RuntimeError("binary_cross_entropy backward called before forward.")
        if self._input is None or self._target is None:
            raise RuntimeError("binary_cross_entropy cached data missing.")

        x = self._input
        t = self._target
        grad = (x - t) / (x * (1 - x))

        if self._weight is not None:
            grad = grad * self._weight

        if self.reduction is None:
            grad = grad * self.result.grad
        elif self.reduction == "sum":
            grad = grad * self.result.grad
        elif self.reduction == "mean":
            grad = grad * (self.result.grad / x.size)

        return grad, None, None


class binary_cross_entropy_with_logits_kernel(Operation):
    def __init__(
        self,
        reduction: str | None = "mean",
        has_weight: bool = True,
        has_pos_weight: bool = True,
    ) -> None:
        super().__init__()
        self.reduction = reduction
        self.has_weight = bool(has_weight)
        self.has_pos_weight = bool(has_pos_weight)

        self._logits = None
        self._target = None
        self._weight = None
        self._pos_weight = None

    def clear(self) -> None:
        super().clear()
        self._logits = None
        self._target = None
        self._weight = None
        self._pos_weight = None

    @func_op(n_in=4, n_ret=1)
    def cpu(
        self, logits: Tensor, target: Tensor, weight: Tensor, pos_weight: Tensor
    ) -> _FuncOpReturnType:
        return self._forward(logits, target, weight, pos_weight, lib_=np, device="cpu")

    @func_op(n_in=4, n_ret=1, device="gpu")
    def gpu(
        self, logits: Tensor, target: Tensor, weight: Tensor, pos_weight: Tensor
    ) -> _FuncOpReturnType:
        return self._forward(logits, target, weight, pos_weight, lib_=mx, device="gpu")

    def _forward(
        self,
        logits: Tensor,
        target: Tensor,
        weight: Tensor,
        pos_weight: Tensor,
        lib_: ModuleType,
        device: _DeviceType,
    ) -> _FuncOpReturnType:
        x = logits.data
        t = target.data

        sp = lib_.log1p(lib_.exp(-lib_.abs(x)))

        if self.has_pos_weight:
            pw = pos_weight.data
            coeff = 1 + (pw - 1) * t
            loss = lib_.maximum(x, 0) - x * t + coeff * sp
        else:
            pw = None
            loss = lib_.maximum(x, 0) - x * t + sp

        if self.has_weight:
            loss = loss * weight.data

        if self.reduction is None:
            out = loss
        elif self.reduction == "sum":
            out = lib_.sum(loss)
        elif self.reduction == "mean":
            out = lib_.mean(loss)
        else:
            raise ValueError("Invalid reduction type. Choose 'mean', 'sum', or 'none'.")

        self._logits = x
        self._target = t
        self._weight = weight.data if self.has_weight else None
        self._pos_weight = pw

        self.result = Tensor(out, device=device)
        return self.result, functools.partial(self.__grad__, lib_=lib_)

    def __grad__(self, lib_: ModuleType) -> _GradType:
        if self.result is None or self.result.grad is None:
            raise RuntimeError(
                "binary_cross_entropy_with_logits backward called before forward."
            )
        if self._logits is None or self._target is None:
            raise RuntimeError("binary_cross_entropy_with_logits cached data missing.")

        x = self._logits
        t = self._target
        sig = 1.0 / (1.0 + lib_.exp(-x))

        if self._pos_weight is not None:
            pw = self._pos_weight
            grad = (sig - t) * (1 + (pw - 1) * t)
        else:
            grad = sig - t

        if self._weight is not None:
            grad = grad * self._weight

        if self.reduction is None:
            grad = grad * self.result.grad
        elif self.reduction == "sum":
            grad = grad * self.result.grad
        elif self.reduction == "mean":
            grad = grad * (self.result.grad / x.size)

        return grad, None, None, None



"""
# `Lucid `

**Lucid** is an educational deep learning framework developed to help users understand
the underlying mechanics of deep learning models and tensor operations.

It is designed to provide a simple yet powerful environment to experiment with neural networks,
optimization, and backpropagation using only `NumPy`.

Lucid is ideal for those who want to learn about the inner workings of deep learning
algorithms and operations without the complexity of high-level frameworks.

[📑 Lucid Documentation](https://chanlumerico.github.io/lucid/build/html/index.html)
"""

from contextlib import contextmanager, AbstractContextManager
from typing import Any, Generator, SupportsIndex, Callable, Self, Optional, Type
from types import TracebackType, ModuleType
from functools import wraps
from pathlib import Path
import inspect

import os
import sys
import json
import math
import numpy as np

_GlobalFlag = bool

USE_CPP_FUNC_OP: _GlobalFlag = False
USE_BAKCWARD_FUSION: _GlobalFlag = True

from lucid._tensor import *
from lucid._func import *
from lucid._utils import *

from lucid._backend.metal import mx

from lucid.types import (
    _ArrayOrScalar,
    _NumPyArray,
    _MLXArray,
    _ArrayLike,
    _ShapeLike,
    _DeviceType,
    _BuiltinNumeric,
    Numeric,
)
from lucid.error import *
from lucid.port import *

import lucid.linalg as linalg
import lucid.random as random
import lucid.einops as einops
import lucid.nn as nn
import lucid.types as types
import lucid.autograd as autograd
import lucid.visual as visual

from lucid._jit.api import compile

_grad_enabled: bool = True
_flops_enabled: bool = False

newaxis = None

pi = math.pi
inf = math.inf

Int = types.Int
Int8, Int16, Int32, Int64 = (types.Int8, types.Int16, types.Int32, types.Int64)
Char, Short, Long = (Int8, Int16, Int64)

Float = types.Float
Float16, Float32, Float64 = (types.Float16, types.Float32, types.Float64)
Half, Double = (Float16, Float64)

Complex = types.Complex
Complex64 = types.Complex64


def tensor(
    data: Tensor | _ArrayOrScalar,
    requires_grad: bool = False,
    keep_grad: bool = False,
    dtype: _BuiltinNumeric | Numeric | None = None,
    device: _DeviceType = "cpu",
) -> Tensor:
    if isinstance(data, Tensor):
        data = data.data
    return Tensor(data, requires_grad, keep_grad, dtype, device)


def to_tensor(
    a: _ArrayLike,
    requires_grad: bool = False,
    keep_grad: bool = False,
    dtype: _BuiltinNumeric | Numeric | None = None,
    device: _DeviceType = "cpu",
) -> Tensor:
    return tensor(a, requires_grad, keep_grad, dtype, device)


class _NoGrad(AbstractContextManager):
    __slots__ = ("_prev_state",)

    def __enter__(self) -> Self:
        global _grad_enabled
        self._prev_state = _grad_enabled

        _grad_enabled = False
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> bool:
        _ = (exc_type, exc_value, traceback)
        global _grad_enabled

        _grad_enabled = self._prev_state
        return False

    def __call__(self, func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            with _NoGrad():
                return func(*args, **kwargs)

        return wrapper


no_grad = _NoGrad


def grad_enabled() -> bool:
    return _grad_enabled


@contextmanager
def count_flops() -> Generator:
    global _flops_enabled
    prev_state = _flops_enabled
    _flops_enabled = True
    try:
        yield
    finally:
        _flops_enabled = prev_state


def flops_enabled() -> bool:
    return _flops_enabled


def shape(a: Tensor | _NumPyArray | _MLXArray) -> _ShapeLike:
    if hasattr(a, "shape"):
        return a.shape
    raise ValueError(f"The argument must be a Tensor or a NumPy array.")


def _check_input_dim(tensor: Tensor, dim: int) -> None:
    if tensor.ndim != dim:
        raise ValueError(f"expected {dim}D input (got {tensor.ndim}D input).")


def _set_tensor_grad(
    tensor: Tensor, grad: _NumPyArray | _MLXArray, at: SupportsIndex = ...
) -> None:
    if not tensor.requires_grad:
        return
    if tensor.grad is None:
        tensor.grad = grad
    else:
        if tensor.is_cpu() and not tensor.grad.flags.writeable:
            tensor.grad = tensor.grad.copy()

        if tensor.is_gpu():
            if at == Ellipsis:
                at = slice(None, None, None)

        if tensor.grad.ndim == 0:
            tensor.grad += grad
        else:
            tensor.grad[at] = tensor.grad[at] + grad


def _check_is_tensor(
    any: Tensor | _ArrayOrScalar,
    device: _DeviceType = "cpu",
    dtype: _BuiltinNumeric | Numeric | None = None,
) -> Tensor:
    if isinstance(any, Tensor):
        return any

    is_scalar = not isinstance(any, (_NumPyArray, _MLXArray, list, tuple))
    if dtype is not None and is_scalar:
        return Tensor(any, device=device, dtype=dtype)

    return Tensor(any, device=device)


def _match_grad_shape(
    data: _NumPyArray | _MLXArray,
    grad: _NumPyArray | _MLXArray,
    device: _DeviceType = "cpu",
) -> _NumPyArray | _MLXArray:
    if data.shape == grad.shape:
        return grad
    if data.ndim == 0:
        return grad.sum()
    lib = np if device == "cpu" else mx
    if grad.ndim == 0:
        return lib.broadcast_to(grad, data.shape)

    data_shape = tuple(data.shape)
    grad_shape = tuple(grad.shape)

    if len(grad_shape) < len(data_shape):
        grad = lib.reshape(
            grad, (1,) * (len(data_shape) - len(grad_shape)) + grad_shape
        )
        grad_shape = tuple(grad.shape)

    aligned_data_shape = data_shape
    if len(data_shape) < len(grad_shape):
        aligned_data_shape = (1,) * (len(grad_shape) - len(data_shape)) + data_shape

    if len(aligned_data_shape) != len(grad_shape):
        raise ValueError("Unknown error occurred.")

    reduce_axes: list[int] = []
    for axis, (d_dim, g_dim) in enumerate(zip(aligned_data_shape, grad_shape)):
        if d_dim == g_dim:
            continue
        if d_dim == 1 and g_dim != 1:
            reduce_axes.append(axis)
        elif g_dim == 1:
            continue
        else:
            raise ValueError(
                f"Cannot reduce grad of {grad.shape} to data of {data.shape}."
            )

    if reduce_axes:
        grad = lib.sum(grad, axis=tuple(reduce_axes), keepdims=True)

    if grad.shape != aligned_data_shape:
        grad = lib.broadcast_to(grad, aligned_data_shape)

    if len(aligned_data_shape) > len(data_shape):
        if grad.size != data.size:
            raise ValueError(
                f"Cannot collapse grad of {grad_shape} to data of {data.shape}."
            )
        return lib.reshape(grad, data_shape)

    return grad


def _get_overloaded_shape(args: int | _ShapeLike) -> _ShapeLike:
    if len(args) == 1 and isinstance(args[0], (tuple, list)):
        shape = tuple(args[0])
    else:
        shape = tuple(args)
    return shape


_PACKAGE_DIR: Path = Path(__file__).resolve().parent
MODELS_REGISTRY_PATH: Path = _PACKAGE_DIR / "models" / "registry.json"

_ModuleReturnFunc = Callable[[Any], nn.Module]
_ModuleClass = type[nn.Module]


def _load_models_registry(path: Path) -> dict[str, Any]:
    if not path.exists():
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "w") as f:
            json.dump({}, f)
        return {}

    try:
        with open(path, "r") as f:
            return json.load(f)
    except json.JSONDecodeError:
        return {}


def _is_legacy_flat_registry(registry: dict[str, Any]) -> bool:
    for value in registry.values():
        if not isinstance(value, dict):
            continue
        if "name" in value and ("param_size" in value or "parameter_size" in value):
            return True
    return False


def _migrate_flat_registry(registry: dict[str, Any]) -> dict[str, Any]:
    nested: dict[str, Any] = {}

    for model_name, info in registry.items():
        if not isinstance(info, dict):
            continue

        category = info.get("category")
        if isinstance(category, str) and category:
            category_parts = [category]
        elif isinstance(category, list) and category:
            category_parts = [str(part) for part in category]
        else:
            category_parts = ["uncategorized"]

        cursor = nested
        for part in category_parts:
            cursor = cursor.setdefault(part, {})

        cursor[model_name] = {
            "parameter_size": int(
                info.get("param_size", info.get("parameter_size", 0))
            ),
            "submodule_count": int(info.get("submodule_count", 0)),
        }

    return nested


def _get_registry_category_path(func: _ModuleReturnFunc) -> list[str]:
    module = sys.modules.get(func.__module__)
    module_file = getattr(module, "__file__", None)
    models_dir = _PACKAGE_DIR / "models"

    if module_file:
        try:
            rel = Path(module_file).resolve().relative_to(models_dir).with_suffix("")
            parts = list(rel.parts)
            if parts and parts[-1] == "__init__":
                parts = parts[:-1]
            if parts:
                return parts
        except ValueError:
            pass

    module_parts = func.__module__.split(".")
    if "models" in module_parts:
        parts = module_parts[module_parts.index("models") + 1 :]
        return parts if parts else ["uncategorized"]
    return ["uncategorized"]


def _has_model_entry(
    registry: dict[str, Any], path: list[str], model_name: str
) -> bool:
    cursor: Any = registry
    for part in path:
        if not isinstance(cursor, dict) or part not in cursor:
            return False
        cursor = cursor[part]
    return isinstance(cursor, dict) and model_name in cursor


def _upsert_model_entry(
    registry: dict[str, Any], path: list[str], model_name: str, entry: dict[str, int]
) -> None:
    cursor = registry
    for part in path:
        cursor = cursor.setdefault(part, {})
    cursor[model_name] = entry


def _build_registry_entry(model: nn.Module, *, is_class: bool) -> dict[str, Any]:
    submodule_count = 0
    for _ in model.modules():
        submodule_count += 1

    entry: dict[str, Any] = {
        "parameter_size": int(model.parameter_size),
        "submodule_count": submodule_count - 1 if submodule_count > 0 else 0,
    }
    if is_class:
        entry["factory_type"] = "class"
    return entry


def _maybe_register_model(target: Any, model: nn.Module, *, is_class: bool) -> None:
    registry = _load_models_registry(MODELS_REGISTRY_PATH)
    if _is_legacy_flat_registry(registry):
        registry = _migrate_flat_registry(registry)

    model_name = target.__name__
    category_path = _get_registry_category_path(target)

    if not _has_model_entry(registry, category_path, model_name):
        _upsert_model_entry(
            registry,
            category_path,
            model_name,
            _build_registry_entry(model, is_class=is_class),
        )
        with open(MODELS_REGISTRY_PATH, "w") as f:
            json.dump(registry, f, indent=4)


def register_model(
    target: _ModuleReturnFunc | _ModuleClass,
) -> _ModuleReturnFunc | _ModuleClass:
    if inspect.isclass(target):
        if not issubclass(target, nn.Module):
            raise TypeError("@register_model class target must inherit from nn.Module.")

        original_init = target.__init__

        @wraps(original_init)
        def wrapped_init(self, *args, **kwargs) -> None:
            weights = kwargs.pop("weights", None)
            original_init(self, *args, **kwargs)

            if os.environ.get("SPHINX_BUILD"):
                return

            _maybe_register_model(target, self, is_class=True)

            if weights is not None:
                import lucid.weights as W

                try:
                    W.apply(self, weights)
                except Exception as e:
                    raise RuntimeError(
                        f"Failed to apply pre-trained weights: {e}"
                    ) from e

        target.__init__ = wrapped_init
        return target

    func = target

    @wraps(func)
    def wrapper(*args, **kwargs) -> nn.Module:
        weights = kwargs.pop("weights", None)

        if os.environ.get("SPHINX_BUILD"):
            return func(*args, **kwargs)

        model = func(*args, **kwargs)
        model._alt_name = func.__name__
        _maybe_register_model(func, model, is_class=False)

        if weights is not None:
            import lucid.weights as W

            try:
                W.apply(model, weights)
            except Exception as e:
                raise RuntimeError(f"Failed to apply pre-trained weights: {e}") from e

        return model

    return wrapper


def _conv_view_limit_mb() -> int:
    from lucid.nn._kernel import conv as _conv_kernel

    return _conv_kernel.get_conv_view_limit_mb()


def __getattr__(name: str) -> Any:
    if name == "CONV_VIEW_LIMIT_MB":
        return _conv_view_limit_mb()
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


def __dir__() -> list[str]:
    return sorted(list(globals().keys()) + ["CONV_VIEW_LIMIT_MB"])


class _LucidModule(ModuleType):
    def __setattr__(self, name: str, value: Any) -> None:
        if name == "CONV_VIEW_LIMIT_MB":
            raise AttributeError(
                "CONV_VIEW_LIMIT_MB is read-only; set LUCID_CONV_VIEW_LIMIT_MB "
                "before importing lucid."
            )
        super().__setattr__(name, value)


if not isinstance(sys.modules[__name__], _LucidModule):
    sys.modules[__name__].__class__ = _LucidModule



from dataclasses import dataclass
import math
from typing import Literal, Sequence

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor

__all__ = ["NCSN", "NCSNConfig"]


@dataclass
class NCSNConfig:
    in_channels: int = 3
    nf: int = 128
    num_classes: int = 10
    dilations: Sequence[int] = (1, 2, 4, 8)
    scale_by_sigma: bool = True
    sigmas: Tensor | Sequence[float] | None = None

    def __post_init__(self) -> None:
        if self.in_channels <= 0:
            raise ValueError("in_channels must be greater than 0")
        if self.nf <= 0:
            raise ValueError("nf must be greater than 0")
        if self.num_classes <= 0:
            raise ValueError("num_classes must be greater than 0")

        self.dilations = tuple(self.dilations)
        if len(self.dilations) != 4 or any(
            dilation <= 0 for dilation in self.dilations
        ):
            raise ValueError("dilations must contain exactly four positive integers")
        if not isinstance(self.scale_by_sigma, bool):
            raise TypeError("scale_by_sigma must be a bool")

        if self.sigmas is not None:
            if isinstance(self.sigmas, Tensor):
                if self.sigmas.ndim != 1:
                    raise ValueError("sigmas must be 1D when provided")
                if self.sigmas.size != self.num_classes:
                    raise ValueError("sigmas length must match num_classes")
            else:
                self.sigmas = tuple(float(sigma) for sigma in self.sigmas)
                if len(self.sigmas) != self.num_classes:
                    raise ValueError("sigmas length must match num_classes")
                if any(sigma <= 0 for sigma in self.sigmas):
                    raise ValueError("sigmas must contain only positive values")


class _CondInstanceNorm(nn.Module):
    def __init__(self, num_features: int, num_classes: int, eps: float = 1e-5) -> None:
        super().__init__()
        self.num_features = num_features
        self.norm = nn.InstanceNorm2d(num_features, affine=False, eps=eps)
        self.embed = nn.Embedding(num_classes, num_features * 2)

        nn.init.constant(self.embed.weight[:, :num_features], 1.0)
        nn.init.constant(self.embed.weight[:, num_features:], 0.0)

    def forward(self, x: Tensor, y: Tensor) -> Tensor:
        if y.dtype != lucid.Long:
            y = y.long()

        h = self.norm(x)
        gamma_beta = self.embed(y)
        gamma, beta = lucid.chunk(gamma_beta, 2, axis=1)

        gamma = gamma.reshape(-1, self.num_features, 1, 1)
        beta = beta.reshape(-1, self.num_features, 1, 1)

        return h * gamma + beta


class _Conv3x3(nn.Module):
    def __init__(
        self, in_channels: int, out_channels: int, dilation: int = 1, bias: bool = True
    ) -> None:
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=3,
            stride=1,
            padding=dilation,
            dilation=dilation,
            bias=bias,
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.conv(x)


class _ResidualConvUnit(nn.Module):
    def __init__(self, channels: int, num_classes: int, dilation: int = 1) -> None:
        super().__init__()
        self.norm1 = _CondInstanceNorm(channels, num_classes)
        self.conv1 = _Conv3x3(channels, channels, dilation=dilation)

        self.norm2 = _CondInstanceNorm(channels, num_classes)
        self.conv2 = _Conv3x3(channels, channels, dilation=dilation)

        self.act = nn.ELU()

    def forward(self, x: Tensor, y: Tensor) -> Tensor:
        h = self.conv1(self.act(self.norm1(x, y)))
        h = self.conv2(self.act(self.norm2(h, y)))

        return x + h


class _RCUBlock(nn.Module):
    def __init__(
        self, channels: int, num_classes: int, num_units: int = 2, dilation: int = 1
    ) -> None:
        super().__init__()
        self.units = nn.ModuleList(
            [
                _ResidualConvUnit(channels, num_classes, dilation=dilation)
                for _ in range(num_units)
            ]
        )

    def forward(self, x: Tensor, y: Tensor) -> Tensor:
        h = x
        for unit in self.units:
            h = unit(h, y)
        return h


class _CondAdapter(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, num_classes: int) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels

        if in_channels == out_channels:
            self.norm = None
            self.conv = None
        else:
            self.norm = _CondInstanceNorm(in_channels, num_classes)
            self.conv = nn.Conv2d(
                in_channels, out_channels, kernel_size=1, stride=1, padding=0
            )
        self.act = nn.ELU()

    def forward(self, x: Tensor, y: Tensor) -> Tensor:
        if self.in_channels == self.out_channels:
            return x
        return self.conv(self.act(self.norm(x, y)))


class _MultiResFusion(nn.Module):
    def __init__(
        self, in_channels_arr: Sequence[int], out_channels: int, num_classes: int
    ) -> None:
        super().__init__()
        self.out_channels = out_channels
        self.norms = nn.ModuleList(
            [_CondInstanceNorm(c, num_classes) for c in in_channels_arr]
        )
        self.convs = nn.ModuleList(
            [
                nn.Conv2d(c, out_channels, kernel_size=3, stride=1, padding=1)
                for c in in_channels_arr
            ]
        )
        self.act = nn.ELU()

    def forward(self, xs: Sequence[Tensor], y: Tensor) -> Tensor:
        if len(xs) != len(self.convs):
            raise ValueError(f"Expected {len(self.convs)} inputs, got {len(xs)}")

        target_h = max(x.shape[-2] for x in xs)
        target_w = max(x.shape[-1] for x in xs)
        fused = None

        for x, norm, conv in zip(xs, self.norms, self.convs):
            h = conv(self.act(norm(x, y)))
            if h.shape[-2:] != (target_h, target_w):
                h = F.interpolate(h, size=(target_h, target_w), mode="nearest")

            fused = h if fused is None else fused + h

        return fused


class _ChainedResPooling(nn.Module):
    def __init__(self, channels: int, num_classes: int, num_stages: int = 4) -> None:
        super().__init__()
        self.norms = nn.ModuleList(
            [_CondInstanceNorm(channels, num_classes) for _ in range(num_stages)]
        )
        self.convs = nn.ModuleList(
            [
                nn.Conv2d(channels, channels, kernel_size=3, stride=1, padding=1)
                for _ in range(num_stages)
            ]
        )
        self.act = nn.ELU()

    def forward(self, x: Tensor, y: Tensor) -> Tensor:
        h = x
        out = x
        for norm, conv in zip(self.norms, self.convs):
            h = self.act(norm(h, y))
            h = F.max_pool2d(h, kernel_size=5, stride=1, padding=2)
            h = conv(h)
            out = out + h

        return out


class _RefineBlock(nn.Module):
    def __init__(
        self, in_channels_arr: Sequence[int], out_channels: int, num_classes: int
    ) -> None:
        super().__init__()
        self.adapters = nn.ModuleList(
            [_CondAdapter(c, out_channels, num_classes) for c in in_channels_arr]
        )
        self.rcu_in = nn.ModuleList(
            [_RCUBlock(out_channels, num_classes, num_units=2) for _ in in_channels_arr]
        )
        self.msf = _MultiResFusion(
            [out_channels] * len(in_channels_arr), out_channels, num_classes
        )
        self.crp = _ChainedResPooling(out_channels, num_classes, num_stages=4)
        self.rcu_out = _RCUBlock(out_channels, num_classes, num_units=2)

    def forward(self, xs: Sequence[Tensor], y: Tensor) -> Tensor:
        if len(xs) != len(self.adapters):
            raise ValueError(f"Expected {len(self.adapters)} inputs, got {len(xs)}")

        hs: list[Tensor] = []
        for x, adapter, rcu in zip(xs, self.adapters, self.rcu_in):
            h = adapter(x, y)
            h = rcu(h, y)
            hs.append(h)

        h = hs[0] if len(hs) == 1 else self.msf(hs, y)
        h = self.crp(h, y)
        h = self.rcu_out(h, y)

        return h


@nn.set_state_dict_pass_attr("sigmas")
class NCSN(nn.Module):
    def __init__(self, config: NCSNConfig) -> None:
        super().__init__()
        self.config = config

        self.in_channels = config.in_channels
        self.nf = config.nf
        self.num_classes = config.num_classes
        self.scale_by_sigma = config.scale_by_sigma

        self.sigmas: nn.Buffer
        self.register_buffer("sigmas", lucid.empty(config.num_classes))

        self.begin_conv = nn.Conv2d(
            config.in_channels, config.nf, kernel_size=3, stride=1, padding=1
        )

        self.stage1 = _RCUBlock(
            config.nf, config.num_classes, num_units=2, dilation=config.dilations[0]
        )
        self.stage2 = _RCUBlock(
            config.nf, config.num_classes, num_units=2, dilation=config.dilations[1]
        )
        self.stage3 = _RCUBlock(
            config.nf, config.num_classes, num_units=2, dilation=config.dilations[2]
        )
        self.stage4 = _RCUBlock(
            config.nf, config.num_classes, num_units=2, dilation=config.dilations[3]
        )

        self.refine4 = _RefineBlock([config.nf], config.nf, config.num_classes)
        self.refine3 = _RefineBlock(
            [config.nf, config.nf], config.nf, config.num_classes
        )
        self.refine2 = _RefineBlock(
            [config.nf, config.nf], config.nf, config.num_classes
        )
        self.refine1 = _RefineBlock(
            [config.nf, config.nf], config.nf, config.num_classes
        )

        self.end_norm = _CondInstanceNorm(config.nf, config.num_classes)
        self.end_act = nn.ELU()
        self.end_conv = nn.Conv2d(
            config.nf, config.in_channels, kernel_size=3, stride=1, padding=1
        )

        self.apply(self._init_weights)
        if config.sigmas is not None:
            sigmas = (
                config.sigmas
                if isinstance(config.sigmas, Tensor)
                else Tensor(config.sigmas, dtype=lucid.Float32)
            )
            self.set_sigmas(sigmas)

    @staticmethod
    def _init_weights(m: nn.Module) -> None:
        if isinstance(m, nn.Conv2d):
            nn.init.xavier_uniform(m.weight)
            if m.bias is not None:
                nn.init.constant(m.bias, 0.0)

    def forward(self, x: Tensor, labels: Tensor) -> Tensor:
        if labels.ndim != 1:
            labels = labels.reshape(-1)

        h = self.begin_conv(x)
        h1 = self.stage1(h, labels)
        h2 = self.stage2(h1, labels)
        h3 = self.stage3(h2, labels)
        h4 = self.stage4(h3, labels)

        r4 = self.refine4([h4], labels)
        r3 = self.refine3([h3, r4], labels)
        r2 = self.refine2([h2, r3], labels)
        r1 = self.refine1([h1, r2], labels)

        out = self.end_conv(self.end_act(self.end_norm(r1, labels)))
        if self.scale_by_sigma:
            if self.sigmas.size != self.num_classes:
                raise RuntimeError(
                    f"'sigmas' buffer has shape {self.sigmas.shape}; "
                    f"expected ({self.num_classes},). Call 'set_sigmas(...)'."
                )

            used_sigmas = self.sigmas[labels].reshape(-1, 1, 1, 1)
            out = out / used_sigmas

        return out

    @lucid.no_grad()
    def set_sigmas(self, sigmas: Tensor) -> None:
        if sigmas.ndim != 1:
            raise ValueError("sigmas must be 1D.")
        if sigmas.size != self.num_classes:
            raise ValueError(
                f"sigmas length ({sigmas.size}) must match "
                f"num_classes ({self.num_classes})."
            )
        tmp = sigmas.detach()
        tmp = tmp.to(self.sigmas.device)
        tmp = tmp.to(self.sigmas.dtype)
        self.sigmas.data = tmp.data

    @staticmethod
    @lucid.no_grad()
    def make_sigmas(sigma_begin: float, sigma_end: float, num_scales: int) -> Tensor:
        if sigma_begin <= 0 or sigma_end <= 0:
            raise ValueError("sigmas must be positive.")
        if sigma_begin <= sigma_end:
            raise ValueError(
                "Expected sigma_begin > sigma_end (descending noise schedule)."
            )
        if num_scales < 2:
            raise ValueError("num_scales must be >= 2.")

        return lucid.exp(
            lucid.linspace(math.log(sigma_begin), math.log(sigma_end), num_scales)
        )

    def get_loss(self, x: Tensor) -> tuple[Tensor, Tensor]:
        batch_size = x.shape[0]
        labels = lucid.random.randint(
            0, self.sigmas.shape[0], (batch_size,), device=x.device
        ).long()
        used_sigmas = self.sigmas[labels].reshape(batch_size, 1, 1, 1)

        noise = lucid.random.randn(x.shape, device=x.device)
        perturbed = x + used_sigmas * noise
        score = self.forward(perturbed, labels)

        loss = lucid.sum((score * used_sigmas + noise) ** 2, axis=(1, 2, 3)).mean()
        return loss, labels

    @lucid.no_grad()
    def sample(
        self,
        n_samples: int,
        image_size: int,
        in_channels: int,
        n_steps_each: int,
        step_lr: float,
        clip: bool = True,
        denoise: bool = False,
        init: Tensor | None = None,
        init_dist: Literal["uniform", "normal"] = "uniform",
        verbose: bool = True,
    ) -> Tensor:
        self.eval()
        if init is None:
            if init_dist == "uniform":
                x = lucid.random.uniform(
                    -1.0,
                    1.0,
                    (n_samples, in_channels, image_size, image_size),
                    device=self.device,
                )
            elif init_dist == "normal":
                x = lucid.random.randn(
                    n_samples, in_channels, image_size, image_size, device=self.device
                )
            else:
                raise ValueError("init_dist must be either 'uniform' or 'normal'.")

        else:
            x = init.to(self.device)
            if x.shape != (n_samples, in_channels, image_size, image_size):
                raise ValueError(
                    f"init has shape {x.shape} but expected "
                    f"{(n_samples, in_channels, image_size, image_size)}."
                )

        from tqdm import tqdm

        total = int(self.sigmas.shape[0]) * int(n_steps_each)
        pbar = (
            tqdm(total=total, desc="Sampling", dynamic_ncols=True) if verbose else None
        )
        for i, sigma in enumerate(self.sigmas):
            labels = lucid.full(n_samples, i, device=self.device, dtype=lucid.Long)
            step_size = step_lr * (sigma / self.sigmas[-1]) ** 2

            for j in range(n_steps_each):
                grad = self.forward(x, labels)
                noise = lucid.random.randn(x.shape, device=self.device)

                x = x + step_size * grad + lucid.sqrt(2.0 * step_size) * noise
                if clip:
                    x = x.clip(-1.0, 1.0)
                if verbose:
                    pbar.update(1)
                    pbar.set_postfix(sigma=f"{sigma.item():.4f}", l=i, t=j)

        if denoise:
            last_label = lucid.full(
                n_samples,
                self.sigmas.shape[0] - 1,
                device=self.device,
                dtype=lucid.Long,
            )
            x = x + (self.sigmas[-1] ** 2) * self.forward(x, last_label)
            if clip:
                x = x.clip(-1.0, 1.0)

        if verbose:
            pbar.close()
        return x
